"""A polygon that can be displayed in 2D space."""
from ladybug_geometry.geometry2d.polygon import Polygon2D
from ladybug.color import Color

from .._base import DISPLAY_MODES
from ._base import _SingleColorBase2D


class DisplayPolygon2D(_SingleColorBase2D):
    """A polygon in 2D space with display properties.

    Args:
        geometry: A ladybug-geometry Polygon2D object.
        color: A ladybug Color object. If None, a default black color will be
            used. (Default: None).
        display_mode: Text to indicate the display mode (surface, wireframe, etc.).
            Choose from the following. (Default: Surface).

            * Surface
            * SurfaceWithEdges
            * Wireframe
            * Points

    Properties:
        * geometry
        * color
        * display_mode
        * vertices
        * segments
        * min
        * max
        * center
        * perimeter
        * area
        * is_clockwise
        * is_convex
        * is_self_intersecting
        * user_data
    """
    __slots__ = ('_display_mode',)

    def __init__(self, geometry, color=None, display_mode='Surface'):
        """Initialize base with shade object."""
        assert isinstance(geometry, Polygon2D), '\
            Expected ladybug_geometry Polygon2D. Got {}'.format(type(geometry))
        _SingleColorBase2D.__init__(self, geometry, color)
        self.display_mode = display_mode

    @classmethod
    def from_dict(cls, data):
        """Initialize a DisplayPolygon2D from a dictionary.

        Args:
            data: A dictionary representation of an DisplayPolygon2D object.
        """
        assert data['type'] == 'DisplayPolygon2D', \
            'Expected DisplayPolygon2D dictionary. Got {}.'.format(data['type'])
        color = Color.from_dict(data['color']) if 'color' in data and data['color'] \
            is not None else None
        d_mode = data['display_mode'] if 'display_mode' in data and \
            data['display_mode'] is not None else 'Surface'
        geo = cls(Polygon2D.from_dict(data['geometry']), color, d_mode)
        if 'user_data' in data and data['user_data'] is not None:
            geo.user_data = data['user_data']
        return geo

    @property
    def display_mode(self):
        """Get or set text to indicate the display mode."""
        return self._display_mode

    @display_mode.setter
    def display_mode(self, value):
        clean_input = value.lower()
        for key in DISPLAY_MODES:
            if key.lower() == clean_input:
                value = key
                break
        else:
            raise ValueError(
                'display_mode {} is not recognized.\nChoose from the '
                'following:\n{}'.format(value, DISPLAY_MODES))
        self._display_mode = value

    @property
    def vertices(self):
        """Get a tuple of Point2Ds that make up the polygon."""
        return self._geometry.vertices

    @property
    def segments(self):
        """Get a tuple of LineSegment2Ds that make up the polygon."""
        return self._geometry.segments

    @property
    def min(self):
        """Get a Point2D for the minimum of the bounding box around the object."""
        return self._geometry.min

    @property
    def max(self):
        """Get a Point2D for the maximum of the bounding box around the object."""
        return self._geometry.max

    @property
    def center(self):
        """Get a Point2D for the center of the bounding box around the object."""
        return self._geometry.center

    @property
    def perimeter(self):
        """Get a number for the perimeter of the polygon."""
        return self._geometry.perimeter

    @property
    def area(self):
        """Get a number for the area of the polygon."""
        return self._geometry.area

    @property
    def is_clockwise(self):
        """Get a boolean for whether the polygon is clockwise."""
        return self._geometry.is_clockwise

    @property
    def is_convex(self):
        """Get a boolean for whether the polygon is convex."""
        return self._geometry.is_convex

    @property
    def is_self_intersecting(self):
        """Get a boolean for whether the polygon is self intersecting."""
        return self._geometry.is_self_intersecting

    def to_dict(self):
        """Return DisplayPolygon2D as a dictionary."""
        base = {'type': 'DisplayPolygon2D'}
        base['geometry'] = self._geometry.to_dict()
        base['color'] = self.color.to_dict()
        base['display_mode'] = self.display_mode
        if self.user_data is not None:
            base['user_data'] = self.user_data
        return base

    def __copy__(self):
        new_g = DisplayPolygon2D(self.geometry, self.color, self.display_mode)
        new_g._user_data = None if self.user_data is None else self.user_data.copy()
        return new_g

    def __repr__(self):
        return 'DisplayPolygon2D: {}'.format(self.geometry)



# coding=utf-8
import pytest

from ladybug_geometry.geometry3d.arc import Arc3D
from ladybug_geometry.geometry3d.plane import Plane
from ladybug_geometry.geometry3d.pointvector import Point3D
from ladybug.color import Color
from ladybug_display.geometry3d.arc import DisplayArc3D
from ladybug_display.altnumber import default

import math


def test_display_arc3d_init():
    """Test the initialization of Arc3D objects and basic properties."""
    grey = Color(100, 100, 100)
    pt = Point3D(2, 0, 2)
    arc = DisplayArc3D(Arc3D(Plane(o=pt), 1, 0, math.pi), grey)
    str(arc)  # test the string representation of the arc

    assert arc.color == grey
    assert arc.line_width == default
    assert arc.line_type == 'Continuous'
    assert arc.c == pt
    assert arc.radius == 1
    assert arc.p1 == Point3D(3, 0, 2)
    assert arc.p2.x == pytest.approx(1, rel=1e-3)
    assert arc.p2.y == pytest.approx(0, rel=1e-3)
    assert arc.p2.z == pytest.approx(2, rel=1e-3)
    assert arc.length == pytest.approx(math.pi, rel=1e-3)
    assert arc.angle == pytest.approx(180, rel=1e-3)
    assert arc.is_circle is False

    blue = Color(0, 0, 100)
    arc.color = blue
    arc.line_width = 2
    arc.line_type = 'Dashed'
    assert arc.color == blue
    assert arc.line_width == 2
    assert arc.line_type == 'Dashed'


def test_arc3_to_from_dict():
    """Test the initialization of Arc3D objects and basic properties."""
    grey = Color(100, 100, 100)
    pt = Point3D(2, 0, 2)
    arc = DisplayArc3D(Arc3D(Plane(o=pt), 1, 0, math.pi), grey)
    arc_dict = arc.to_dict()
    new_arc = DisplayArc3D.from_dict(arc_dict)
    assert isinstance(new_arc, DisplayArc3D)
    assert new_arc.to_dict() == arc_dict



# coding=utf-8
import pytest

from ladybug_geometry.geometry2d import Point2D, Vector2D
from ladybug.color import Color
from ladybug_display.geometry2d.point import DisplayPoint2D


def test_display_point2d_init():
    """Test the initialization of DisplayPoint2D objects and basic properties."""
    grey = Color(100, 100, 100)
    pt = DisplayPoint2D(Point2D(0, 2), grey)
    str(pt)  # test the string representation of the point

    assert pt.x == 0
    assert pt.y == 2
    assert pt.color == grey
    assert pt[0] == 0
    assert pt[1] == 2

    pt_tuple = tuple(i for i in pt)
    assert pt_tuple == (0, 2)


def test_display_point2d_to_from_dict():
    """Test the initialization of DisplayPoint2D objects and basic properties."""
    grey = Color(100, 100, 100)
    pt = DisplayPoint2D(Point2D(0, 2), grey)
    pt_dict = pt.to_dict()
    new_pt = DisplayPoint2D.from_dict(pt_dict)
    assert isinstance(new_pt, DisplayPoint2D)
    assert new_pt.to_dict() == pt_dict


def test_move():
    """Test the DisplayPoint2D move method."""
    grey = Color(100, 100, 100)
    pt_1 = DisplayPoint2D(Point2D(2, 2), grey)
    vec_1 = Vector2D(0, 2)
    pt_1.move(vec_1)
    assert pt_1.geometry == Point2D(2, 4)


def test_scale():
    """Test the DisplayPoint2D scale method."""
    grey = Color(100, 100, 100)
    pt_1 = DisplayPoint2D(Point2D(2, 2), grey)
    origin_1 = Point2D(0, 2)
    pt_1.scale(2, origin_1)
    assert pt_1.geometry == Point2D(4, 2)


def test_scale_world_origin():
    """Test the DisplayPoint2D scale method with None origin."""
    grey = Color(100, 100, 100)
    pt_1 = DisplayPoint2D(Point2D(2, 2), grey)
    pt_2 = DisplayPoint2D(Point2D(-2, -2), grey)
    pt_1.scale(2)
    assert pt_1.geometry == Point2D(4, 4)
    pt_2.scale(-2)
    assert pt_2.geometry == Point2D(4, 4)


def test_rotate():
    """Test the Point2D rotate_xy method."""
    grey = Color(100, 100, 100)
    pt_1 = DisplayPoint2D(Point2D(2, 2), grey)
    origin_1 = Point2D(0, 2)

    pt_1.rotate(180, origin_1)
    assert pt_1.x == pytest.approx(-2, rel=1e-3)
    assert pt_1.y == pytest.approx(2, rel=1e-3)


def test_reflect():
    """Test the Point2D reflect method."""
    grey = Color(100, 100, 100)
    pt_1 = DisplayPoint2D(Point2D(2, 2), grey)
    origin_1 = Point2D(0, 1)
    normal_1 = Vector2D(0, 1)

    pt_1.reflect(normal_1, origin_1)
    assert pt_1.geometry == Point2D(2, 0)



"""An arc that can be displayed in 3D space."""
import math

from ladybug_geometry.geometry3d.arc import Arc3D
from ladybug.color import Color

from ladybug_display.altnumber import default
from ._base import _LineCurveBase3D


class DisplayArc3D(_LineCurveBase3D):
    """An arc in 3D space with display properties.

    Args:
        geometry: A ladybug-geometry Arc3D object.
        color: A ladybug Color object. If None, a default black color will be
            used. (Default: None).
        line_width: Number for line width in pixels (for the screen) or  millimeters
            (in print). This can also be the Default object to indicate that the
            default settings of the interface should be used.
        line_type: Get or set text to indicate the type of line to display. 
            Choose from the following. (Default: "Continuous")

            * Continuous
            * Dashed
            * Dotted
            * DashDot
            * Center
            * Border
            * Hidden

    Properties:
        * geometry
        * color
        * line_width
        * line_type
        * plane
        * radius
        * p1
        * p2
        * c
        * angle
        * length
        * is_circle
        * user_data
    """
    __slots__ = ()

    def __init__(self, geometry, color=None, line_width=default, line_type='Continuous'):
        """Initialize base with shade object."""
        assert isinstance(geometry, Arc3D), '\
            Expected ladybug_geometry Arc3D. Got {}'.format(type(geometry))
        _LineCurveBase3D.__init__(self, geometry, color, line_width, line_type)

    @classmethod
    def from_dict(cls, data):
        """Initialize a DisplayArc3D from a dictionary.

        Args:
            data: A dictionary representation of an DisplayArc3D object.
        """
        assert data['type'] == 'DisplayArc3D', \
            'Expected DisplayArc3D dictionary. Got {}.'.format(data['type'])
        color = Color.from_dict(data['color']) if 'color' in data and data['color'] \
            is not None else None
        lw = default if 'line_width' not in data or \
            data['line_width'] == default.to_dict() else data['line_width']
        lt = data['line_type'] if 'line_type' in data and data['line_type'] \
            is not None else 'Continuous'
        geo = cls(Arc3D.from_dict(data['geometry']), color, lw, lt)
        if 'user_data' in data and data['user_data'] is not None:
            geo.user_data = data['user_data']
        return geo

    @property
    def plane(self):
        """Get a Plane which the arc lies with an origin representing the center."""
        return self._geometry.plane

    @property
    def radius(self):
        """Get a number for the radius of the arc."""
        return self._geometry.radius

    @property
    def p1(self):
        """Get a Point3D representing the first end point of the arc."""
        return self._geometry.p1

    @property
    def p2(self):
        """Get a Point3D representing the second end point of the arc."""
        return self._geometry.p2

    @property
    def c(self):
        """Get a Point3D representing the center of the arc."""
        return self._geometry.c

    @property
    def angle(self):
        """The total angle over the domain of the arc in degrees."""
        return math.degrees(self._geometry.angle)

    @property
    def length(self):
        """Get a number for the length of the arc."""
        return self._geometry.length

    @property
    def is_circle(self):
        """Get a boolean for whether the arc is a circle."""
        return self._geometry.is_circle

    def to_dict(self):
        """Return DisplayArc3D as a dictionary."""
        base = {'type': 'DisplayArc3D'}
        base['geometry'] = self._geometry.to_dict()
        base['color'] = self.color.to_dict()
        base['line_width'] = default.to_dict() if \
            self.line_width == default else self.line_width
        base['line_type'] = self.line_type
        if self.user_data is not None:
            base['user_data'] = self.user_data
        return base

    def __copy__(self):
        new_g = DisplayArc3D(
            self.geometry, self.color, self.line_width, self.line_type)
        new_g._user_data = None if self.user_data is None else self.user_data.copy()
        return new_g

    def __repr__(self):
        return 'DisplayArc3D: {}'.format(self.geometry)



# coding=utf-8
import pytest

from ladybug_geometry.geometry3d import Point3D, Vector3D
from ladybug_geometry.geometry3d.plane import Plane
from ladybug.color import Color
from ladybug_display.geometry3d.point import DisplayPoint3D


def test_display_point3d_init():
    """Test the initialization of DisplayPoint3D objects and basic properties."""
    grey = Color(100, 100, 100)
    pt = DisplayPoint3D(Point3D(0, 2, 0), grey)
    str(pt)  # test the string representation of the point

    assert pt.x == 0
    assert pt.y == 2
    assert pt.z == 0
    assert pt.color == grey
    assert pt[0] == 0
    assert pt[1] == 2
    assert pt[2] == 0

    pt_tuple = tuple(i for i in pt)
    assert pt_tuple == (0, 2, 0)


def test_display_point3d_to_from_dict():
    """Test the initialization of DisplayPoint3D objects and basic properties."""
    grey = Color(100, 100, 100)
    pt = DisplayPoint3D(Point3D(0, 2, 0), grey)
    pt_dict = pt.to_dict()
    new_pt = DisplayPoint3D.from_dict(pt_dict)
    assert isinstance(new_pt, DisplayPoint3D)
    assert new_pt.to_dict() == pt_dict


def test_move():
    """Test the DisplayPoint3D move method."""
    grey = Color(100, 100, 100)
    pt_1 = DisplayPoint3D(Point3D(2, 2, 0), grey)
    vec_1 = Vector3D(0, 2, 2)
    pt_1.move(vec_1)
    assert pt_1.geometry == Point3D(2, 4, 2)


def test_scale():
    """Test the DisplayPoint3D scale method."""
    grey = Color(100, 100, 100)
    pt_1 = DisplayPoint3D(Point3D(2, 2, 2), grey)
    origin_1 = Point3D(0, 2, 2)
    pt_1.scale(2, origin_1)
    assert pt_1.geometry == Point3D(4, 2, 2)


def test_scale_world_origin():
    """Test the DisplayPoint3D scale method with None origin."""
    grey = Color(100, 100, 100)
    pt_1 = DisplayPoint3D(Point3D(2, 2, 2), grey)
    pt_2 = DisplayPoint3D(Point3D(-2, -2, -2), grey)
    pt_1.scale(2)
    assert pt_1.geometry == Point3D(4, 4, 4)
    pt_2.scale(-2)
    assert pt_2.geometry == Point3D(4, 4, 4)


def test_rotate():
    """Test the DisplayPoint3D rotate method."""
    grey = Color(100, 100, 100)
    pt_1 = DisplayPoint3D(Point3D(2, 2, 2), grey)
    axis_1 = Vector3D(-1, 0, 0)
    origin_1 = Point3D(0, 2, 0)

    pt_1.rotate(axis_1, 180, origin_1)
    assert pt_1.x == pytest.approx(2, rel=1e-3)
    assert pt_1.y == pytest.approx(2, rel=1e-3)
    assert pt_1.z == pytest.approx(-2, rel=1e-3)


def test_rotate_xy():
    """Test the Point3D rotate_xy method."""
    grey = Color(100, 100, 100)
    pt_1 = DisplayPoint3D(Point3D(2, 2, -2), grey)
    origin_1 = Point3D(0, 2, 0)

    pt_1.rotate_xy(180, origin_1)
    assert pt_1.x == pytest.approx(-2, rel=1e-3)
    assert pt_1.y == pytest.approx(2, rel=1e-3)
    assert pt_1.z == -2


def test_reflect():
    """Test the Point3D reflect method."""
    grey = Color(100, 100, 100)
    pt_1 = DisplayPoint3D(Point3D(2, 2, 2), grey)
    origin_1 = Point3D(0, 1, 0)
    normal_1 = Vector3D(0, 1, 0)
    plane = Plane(normal_1, origin_1)

    pt_1.reflect(plane)
    assert pt_1.geometry == Point3D(2, 0, 2)



# coding=utf-8
import pytest

from ladybug_geometry.geometry3d import Vector3D
from ladybug.color import Color
from ladybug_display.geometry3d.vector import DisplayVector3D


def test_display_vector3d_init():
    """Test the initialization of DisplayVector3D objects and basic properties."""
    grey = Color(100, 100, 100)
    vec = DisplayVector3D(Vector3D(0, 2, 0), grey)
    str(vec)  # test the string representation of the vector

    assert vec.x == 0
    assert vec.y == 2
    assert vec.z == 0
    assert vec.color == grey
    assert vec[0] == 0
    assert vec[1] == 2
    assert vec[2] == 0
    assert vec.magnitude == 2

    pt_tuple = tuple(i for i in vec)
    assert pt_tuple == (0, 2, 0)

    vec.normalize()
    assert vec.x == 0
    assert vec.z == 0
    assert vec.magnitude == 1

    norm_vec = vec * 3
    assert norm_vec.magnitude == 3


def test_display_vector3d_to_from_dict():
    """Test the initialization of DisplayVector3D objects and basic properties."""
    grey = Color(100, 100, 100)
    vec = DisplayVector3D(Vector3D(0, 2, 0), grey)
    vec_dict = vec.to_dict()
    new_vec = DisplayVector3D.from_dict(vec_dict)
    assert isinstance(new_vec, DisplayVector3D)
    assert new_vec.to_dict() == vec_dict


def test_display_vector3d_angle():
    """Test the methods that get the angle between DisplayVector3D objects."""
    grey = Color(100, 100, 100)
    vec_1 = DisplayVector3D(Vector3D(0, 2, 0), grey)
    vec_2 = DisplayVector3D(Vector3D(2, 0, 0), grey)
    vec_3 = DisplayVector3D(Vector3D(0, -2, 0), grey)
    vec_4 = DisplayVector3D(Vector3D(-2, 0, 0), grey)
    vec_5 = DisplayVector3D(Vector3D(0, 0, 2), grey)
    vec_6 = DisplayVector3D(Vector3D(0, 0, -2), grey)
    assert vec_1.angle(vec_2) == pytest.approx(90, rel=1e-3)
    assert vec_1.angle(vec_3) == pytest.approx(180, rel=1e-3)
    assert vec_1.angle(vec_4) == pytest.approx(90, rel=1e-3)
    assert vec_1.angle(vec_5) == pytest.approx(90, rel=1e-3)
    assert vec_5.angle(vec_6) == pytest.approx(180, rel=1e-3)
    assert vec_1.angle(vec_1) == pytest.approx(0, rel=1e-3)



"""A vector that can be displayed in 2D space."""
import math

from ladybug_geometry.geometry2d.pointvector import Vector2D
from ladybug.color import Color

from .._base import _DisplayBase


class DisplayVector2D(_DisplayBase):
    """A vector in 2D space with display properties.

    Args:
        geometry: A ladybug-geometry Vector2D object.
        color: A ladybug Color object. If None, a default black color will be
            used. (Default: None).

    Properties:
        * geometry
        * color
        * x
        * y
        * magnitude
        * user_data
    """
    __slots__ = ('_color',)

    def __init__(self, geometry, color=None):
        """Initialize base with shade object."""
        assert isinstance(geometry, Vector2D), '\
            Expected ladybug_geometry Vector2D. Got {}'.format(type(geometry))
        _DisplayBase.__init__(self, geometry)
        self.color = color

    @classmethod
    def from_dict(cls, data):
        """Initialize a DisplayVector2D from a dictionary.

        Args:
            data: A dictionary representation of an DisplayVector2D object.
        """
        assert data['type'] == 'DisplayVector2D', \
            'Expected DisplayVector2D dictionary. Got {}.'.format(data['type'])
        color = Color.from_dict(data['color']) if 'color' in data and data['color'] \
            is not None else None
        geo = cls(Vector2D.from_dict(data['geometry']), color)
        if 'user_data' in data and data['user_data'] is not None:
            geo.user_data = data['user_data']
        return geo

    @property
    def color(self):
        """Get or set a color for this object."""
        return self._color

    @color.setter
    def color(self, value):
        if value is None:
            value = Color(0, 0, 0)
        else:
            assert isinstance(value, Color), 'Expected Color for ladybug_display ' \
                'object color. Got {}.'.format(type(value))
        self._color = value

    @property
    def x(self):
        """Get the X coordinate."""
        return self._geometry.x

    @property
    def y(self):
        """Get the Y coordinate."""
        return self._geometry.y

    @property
    def magnitude(self):
        """Get the magnitude of the vector."""
        return self._geometry.magnitude

    def normalize(self):
        """Ensure that this DisplayVector2D is a unit vector (magnitude=1)."""
        self._geometry = self._geometry.normalize()

    def reverse(self):
        """Get a copy of this vector that is reversed."""
        self._geometry = self._geometry.__neg__()

    def dot(self, other):
        """Get the dot product of this vector with another DisplayVector2D."""
        return self.geometry.dot(other.geometry)

    def angle(self, other):
        """Get the angle between this vector and another DisplayVector2D in degrees."""
        return math.degrees(self.geometry.angle(other.geometry))

    def rotate(self, angle):
        """Rotate this geometry by a certain angle around an axis and origin.

        Args:
            axis: A ladybug_geometry Vector2D axis representing the axis of rotation.
            angle: An angle for rotation in degrees.
        """
        self._geometry = self.geometry.rotate(math.radians(angle))

    def reflect(self, normal):
        """Reflect this geometry across a plane with the input normal vector.

        Args:
            normal: A Vector2D representing the normal vector for the plane across
                which the vector will be reflected. THIS VECTOR MUST BE NORMALIZED.
        """
        self._geometry = self.geometry.reflect(normal)

    def to_dict(self):
        """Return DisplayVector2D as a dictionary.
        """
        base = {'type': 'DisplayVector2D'}
        base['geometry'] = self._geometry.to_dict()
        base['color'] = self.color.to_dict()
        if self.user_data is not None:
            base['user_data'] = self.user_data
        return base

    def __copy__(self):
        new_g = DisplayVector2D(self.geometry, self.color)
        new_g._user_data = None if self.user_data is None else self.user_data.copy()
        return new_g

    def __getitem__(self, key):
        return (self.x, self.y)[key]

    def __iter__(self):
        return iter((self.x, self.y))

    def __mul__(self, other):
        assert type(other) in (int, float), 'Cannot multiply types {} and {}'.format(
            self.__class__.__name__, type(other))
        new_g = DisplayVector2D(self.geometry * other, self.color)
        new_g._user_data = None if self.user_data is None else self.user_data.copy()
        return new_g

    def __div__(self, other):
        assert type(other) in (int, float), 'Cannot divide types {} and {}'.format(
            self.__class__.__name__, type(other))
        new_g = DisplayVector2D(self.geometry / other, self.color)
        new_g._user_data = None if self.user_data is None else self.user_data.copy()
        return new_g

    def __repr__(self):
        return 'DisplayVector2D: {}'.format(self.geometry)



"""A line segment that can be displayed in 3D space."""
from ladybug_geometry.geometry3d.line import LineSegment3D
from ladybug.color import Color

from ladybug_display.altnumber import default
from ._base import _LineCurveBase3D


class DisplayLineSegment3D(_LineCurveBase3D):
    """A single line segment in 3D space with display properties.

    Args:
        geometry: A ladybug-geometry LineSegment3D object.
        color: A ladybug Color object. If None, a default black color will be
            used. (Default: None).
        line_width: Number for line width in pixels (for the screen) or  millimeters
            (in print). This can also be the Default object to indicate that the
            default settings of the interface should be used.
        line_type: Get or set text to indicate the type of line to display. 
            Choose from the following. (Default: "Continuous")

            * Continuous
            * Dashed
            * Dotted
            * DashDot
            * Center
            * Border
            * Hidden

    Properties:
        * geometry
        * color
        * line_width
        * line_type
        * p
        * v
        * p1
        * p2
        * length
        * user_data
    """
    __slots__ = ()

    def __init__(self, geometry, color=None, line_width=default, line_type='Continuous'):
        """Initialize base with shade object."""
        assert isinstance(geometry, LineSegment3D), '\
            Expected ladybug_geometry LineSegment3D. Got {}'.format(type(geometry))
        _LineCurveBase3D.__init__(self, geometry, color, line_width, line_type)

    @classmethod
    def from_dict(cls, data):
        """Initialize a DisplayLineSegment3D from a dictionary.

        Args:
            data: A dictionary representation of an DisplayLineSegment3D object.
        """
        assert data['type'] == 'DisplayLineSegment3D', \
            'Expected DisplayLineSegment3D dictionary. Got {}.'.format(data['type'])
        color = Color.from_dict(data['color']) if 'color' in data and data['color'] \
            is not None else None
        lw = default if 'line_width' not in data or \
            data['line_width'] == default.to_dict() else data['line_width']
        lt = data['line_type'] if 'line_type' in data and data['line_type'] \
            is not None else 'Continuous'
        geo = cls(LineSegment3D.from_dict(data['geometry']), color, lw, lt)
        if 'user_data' in data and data['user_data'] is not None:
            geo.user_data = data['user_data']
        return geo

    @property
    def p(self):
        """Get a Point3D representing the base of the line segment."""
        return self._geometry.p

    @property
    def v(self):
        """Get a Vector3D representing the direction of the line segment."""
        return self._geometry.v

    @property
    def p1(self):
        """Get a Point3D representing the first point of the line segment."""
        return self._geometry.p1

    @property
    def p2(self):
        """Get a Point3D representing the second point of the line segment."""
        return self._geometry.p2

    @property
    def length(self):
        """Get a number for the length of the line segment."""
        return self._geometry.length

    def to_dict(self):
        """Return DisplayLineSegment3D as a dictionary."""
        base = {'type': 'DisplayLineSegment3D'}
        base['geometry'] = self._geometry.to_dict()
        base['color'] = self.color.to_dict()
        base['line_width'] = default.to_dict() if \
            self.line_width == default else self.line_width
        base['line_type'] = self.line_type
        if self.user_data is not None:
            base['user_data'] = self.user_data
        return base

    def __copy__(self):
        new_g = DisplayLineSegment3D(
            self.geometry, self.color, self.line_width, self.line_type)
        new_g._user_data = None if self.user_data is None else self.user_data.copy()
        return new_g

    def __repr__(self):
        return 'DisplayLineSegment3D: {}'.format(self.geometry)



"""A ray that can be displayed in 3D space."""
from ladybug_geometry.geometry3d.ray import Ray3D
from ladybug.color import Color

from ._base import _SingleColorBase3D


class DisplayRay3D(_SingleColorBase3D):
    """A ray in 3D space with display properties.

    Args:
        geometry: A ladybug-geometry Ray3D object.
        color: A ladybug Color object. If None, a default black color will be
            used. (Default: None).

    Properties:
        * geometry
        * color
        * p
        * v
        * user_data
    """
    __slots__ = ()

    def __init__(self, geometry, color=None):
        """Initialize base with shade object."""
        assert isinstance(geometry, Ray3D), '\
            Expected ladybug_geometry Ray3D. Got {}'.format(type(geometry))
        _SingleColorBase3D.__init__(self, geometry, color)

    @classmethod
    def from_dict(cls, data):
        """Initialize a DisplayRay3D from a dictionary.

        Args:
            data: A dictionary representation of an DisplayRay3D object.
        """
        assert data['type'] == 'DisplayRay3D', \
            'Expected DisplayRay3D dictionary. Got {}.'.format(data['type'])
        color = Color.from_dict(data['color']) if 'color' in data and data['color'] \
            is not None else None
        geo = cls(Ray3D.from_dict(data['geometry']), color)
        if 'user_data' in data and data['user_data'] is not None:
            geo.user_data = data['user_data']
        return geo

    @property
    def p(self):
        """Get a Point3D representing the base of the ray."""
        return self._geometry.p

    @property
    def v(self):
        """Get a Vector3D representing the direction of the ray."""
        return self._geometry.v
    
    def reverse(self):
        """Reverse this DisplayRay3D."""
        self._geometry = self._geometry.reverse()

    def to_dict(self):
        """Return DisplayRay3D as a dictionary."""
        base = {'type': 'DisplayRay3D'}
        base['geometry'] = self._geometry.to_dict()
        base['color'] = self.color.to_dict()
        if self.user_data is not None:
            base['user_data'] = self.user_data
        return base

    def __copy__(self):
        new_g = DisplayRay3D(self.geometry, self.color)
        new_g._user_data = None if self.user_data is None else self.user_data.copy()
        return new_g

    def __repr__(self):
        return 'DisplayRay3D: {}'.format(self.geometry)



# coding=utf-8
from ladybug_geometry.geometry3d.pointvector import Point3D
from ladybug_geometry.geometry3d.plane import Plane
from ladybug_geometry.geometry2d.mesh import Mesh2D
from ladybug_geometry.geometry3d.mesh import Mesh3D
from ladybug_geometry.geometry3d.polyface import Polyface3D

from ladybug.graphic import GraphicContainer
from ladybug.legend import Legend, LegendParameters

from ladybug_display.visualization import VisualizationSet, ContextGeometry, \
    AnalysisGeometry, VisualizationData


def test_init_visualization_set():
    """Test the initialization of VisualizationSet objects."""
    con_geo = Polyface3D.from_box(2, 4, 2, base_plane=Plane(o=Point3D(0, 2, 0)))
    context = ContextGeometry('Building_Massing', [con_geo])
    mesh2d = Mesh2D.from_grid(num_x=2, num_y=2)
    mesh3d = Mesh3D.from_mesh2d(mesh2d)
    data = VisualizationData([0, 1, 2, 3])
    a_geo = AnalysisGeometry('Test_Results', [mesh3d], [data])
    vis_set = VisualizationSet('Test_Set', [a_geo, context])
    str(vis_set)

    assert a_geo.matching_method == 'faces'
    assert len(data) == 4
    assert data[0] == 0
    assert data[-1] == 3
    for item in data:
        assert isinstance(item, (float, int))

    graphic_con = vis_set.graphic_container()
    assert isinstance(graphic_con, GraphicContainer)
    assert len(graphic_con.values) == 4
    assert isinstance(graphic_con.legend, Legend)
    assert graphic_con.value_colors == graphic_con.legend.value_colors

    assert graphic_con.legend_parameters.is_base_plane_default
    assert graphic_con.legend_parameters.is_segment_height_default
    assert graphic_con.legend_parameters.is_segment_width_default
    assert graphic_con.legend_parameters.is_text_height_default
    assert graphic_con.legend_parameters.base_plane != Plane()

    assert isinstance(graphic_con.lower_title_location, Plane)
    assert isinstance(graphic_con.upper_title_location, Plane)
    assert graphic_con.lower_title_location != Plane()
    assert graphic_con.upper_title_location != Plane()


def test_init_visualization_set_legend_parameters():
    """Test the initialization of VisualizationSet objects with a LegendParameters."""
    con_geo = Polyface3D.from_box(2, 4, 2, base_plane=Plane(o=Point3D(0, 2, 0)))
    context = ContextGeometry('Building_Massing', [con_geo])
    mesh2d = Mesh2D.from_grid(num_x=2, num_y=2)
    mesh3d = Mesh3D.from_mesh2d(mesh2d)
    legend_par = LegendParameters(base_plane=Plane(o=Point3D(2, 2, 0)))
    legend_par.vertical = False
    legend_par.segment_height = 0.25
    legend_par.segment_width = 0.5
    legend_par.text_height = 0.15
    data = VisualizationData([-1, 0, 1, 2], legend_par)
    a_geo = AnalysisGeometry('Test_Results', [mesh3d], [data])
    vis_set = VisualizationSet('Test_Set', [a_geo, context])

    graphic_con = vis_set.graphic_container()
    assert not graphic_con.legend_parameters.is_base_plane_default
    assert not graphic_con.legend_parameters.is_segment_height_default
    assert not graphic_con.legend_parameters.is_segment_width_default
    assert not graphic_con.legend_parameters.is_text_height_default
    assert not graphic_con.legend_parameters.vertical
    assert graphic_con.legend_parameters.base_plane.o == Point3D(2, 2, 0)
    assert graphic_con.legend_parameters.segment_height == 0.25
    assert graphic_con.legend_parameters.segment_width == 0.5
    assert graphic_con.legend_parameters.text_height == 0.15


def test_to_from_dict():
    """Test the to/from dict methods."""
    con_geo = Polyface3D.from_box(2, 4, 2, base_plane=Plane(o=Point3D(0, 2, 0)))
    context = ContextGeometry('Building_Massing', [con_geo])
    mesh2d = Mesh2D.from_grid(num_x=2, num_y=2)
    mesh3d = Mesh3D.from_mesh2d(mesh2d)
    data = VisualizationData([0, 1, 2, 3])
    a_geo = AnalysisGeometry('Test_Results', [mesh3d], [data])
    vis_set = VisualizationSet('Test_Set', [a_geo, context])

    vis_set_dict = vis_set.to_dict()
    new_vis_set = VisualizationSet.from_dict(vis_set_dict)
    assert new_vis_set.to_dict() == vis_set_dict



# coding: utf-8
"""Base classes for all 2D geometry objects."""
import math

from ladybug.color import Color

from ladybug_display._base import _DisplayBase, LINE_TYPES
from ladybug_display.altnumber import default
from ladybug_display.typing import float_positive


class _DisplayBase2D(_DisplayBase):
    """A base class for all 2D ladybug-display geometry objects.

    Args:
        geometry: A ladybug-geometry object.

    Properties:
        * geometry
        * user_data
    """
    __slots__ = ()

    def __init__(self, geometry):
        """Initialize base with shade object."""
        _DisplayBase.__init__(self, geometry)

    def move(self, moving_vec):
        """Move this geometry along a vector.

        Args:
            moving_vec: A ladybug_geometry Vector with the direction and distance
                to move the geometry.
        """
        self._geometry = self.geometry.move(moving_vec)

    def rotate(self, angle, origin):
        """Rotate this geometry counterclockwise by a certain angle.

        Args:
            angle: An angle for rotation in degrees.
            origin: A Point2D for the origin around which the line segment will
                be rotated.
        """
        self._geometry = self.geometry.rotate(math.radians(angle), origin)

    def reflect(self, normal, origin):
        """Reflect this geometry across a plane defined by a normal and origin.

        Args:
            normal: A Vector2D representing the normal vector for the plane across
                which the line segment will be reflected. THIS VECTOR MUST BE NORMALIZED.
            origin: A Point2D representing the origin from which to reflect.
        """
        self._geometry = self.geometry.reflect(normal, origin)

    def scale(self, factor, origin=None):
        """Scale this geometry by a factor from an origin point.

        Args:
            factor: A number representing how much the object should be scaled.
            origin: A ladybug_geometry Point representing the origin from which
                to scale. If None, it will be scaled from the World origin.
        """
        self._geometry = self.geometry.scale(factor, origin)

    def __repr__(self):
        return 'Ladybug Display 2D Base Object'


class _SingleColorBase2D(_DisplayBase2D):
    """A base class for ladybug-display geometry objects with a single color.

    Args:
        geometry: A ladybug-geometry object.
        color: A ladybug Color object. If None, a default black color will be
            used. (Default: None).

    Properties:
        * geometry
        * color
        * user_data
    """
    __slots__ = ('_color',)

    def __init__(self, geometry, color=None):
        """Initialize base with shade object."""
        _DisplayBase2D.__init__(self, geometry)
        self.color = color

    @property
    def color(self):
        """Get or set a color for this object."""
        return self._color

    @color.setter
    def color(self, value):
        if value is None:
            value = Color(0, 0, 0)
        else:
            assert isinstance(value, Color), 'Expected Color for ladybug_display ' \
                'object color. Got {}.'.format(type(value))
        self._color = value


class _LineCurveBase2D(_SingleColorBase2D):
    """A base class for all line-like 2D geometry objects.

    Args:
        geometry: A ladybug-geometry object.
        color: A ladybug Color object. If None, a default black color will be
            used. (Default: None).
        line_width: Number for line width in pixels (for the screen) or millimeters
            (in print). This can also be the Default object to indicate that the
            default settings of the interface should be used.
        line_type: Get or set text to indicate the type of line to display.
            Choose from the following. (Default: "Continuous")

            * Continuous
            * Dashed
            * Dotted
            * DashDot
            * Center
            * Border
            * Hidden

    Properties:
        * geometry
        * color
        * line_width
        * line_type
        * user_data
    """
    __slots__ = ('_line_width', '_line_type')

    def __init__(self, geometry, color=None, line_width=default, line_type='Continuous'):
        """Initialize base with shade object."""
        _SingleColorBase2D.__init__(self, geometry, color)
        self.line_width = line_width
        self.line_type = line_type

    @property
    def line_width(self):
        """Get or set a color for this object."""
        return self._line_width

    @line_width.setter
    def line_width(self, value):
        if value == default:
            self._line_width = default
        else:
            self._line_width = float_positive(value, 'line width')

    @property
    def line_type(self):
        """Get or set text to indicate the type of line to display."""
        return self._line_type

    @line_type.setter
    def line_type(self, value):
        clean_input = value.lower()
        for key in LINE_TYPES:
            if key.lower() == clean_input:
                value = key
                break
        else:
            raise ValueError(
                'line_type {} is not recognized.\nChoose from the '
                'following:\n{}'.format(value, LINE_TYPES))
        self._line_type = value



# coding=utf-8
from ladybug_geometry.geometry3d.pointvector import Point3D
from ladybug_geometry.geometry3d.polyface import Polyface3D
from ladybug.color import Color
from ladybug_display.geometry3d.polyface import DisplayPolyface3D


def test_polyface3d_init_solid():
    """Test the initialization of Polyface3D and basic properties of solid objects."""
    grey = Color(100, 100, 100)
    pts = [Point3D(0, 0, 0), Point3D(0, 2, 0), Point3D(2, 2, 0), Point3D(2, 0, 0),
           Point3D(0, 0, 2), Point3D(0, 2, 2), Point3D(2, 2, 2), Point3D(2, 0, 2)]
    face_indices = [[(0, 1, 2, 3)], [(0, 4, 5, 1)], [(0, 3, 7, 4)],
                    [(2, 1, 5, 6)], [(2, 3, 7, 6)], [(4, 5, 6, 7)]]
    polyface = DisplayPolyface3D(Polyface3D(pts, face_indices), [grey])

    assert polyface.colors == (grey,)
    assert polyface.display_mode == 'Surface'
    assert len(polyface.vertices) == 8
    assert len(polyface.face_indices) == 6
    assert len(polyface.faces) == 6
    assert len(polyface.edge_indices) == 12
    assert len(polyface.edges) == 12
    assert len(polyface.naked_edges) == 0
    assert len(polyface.non_manifold_edges) == 0
    assert len(polyface.internal_edges) == 12
    assert polyface.area == 24
    assert polyface.volume == 8
    assert polyface.is_solid
    for face in polyface.faces:
        assert face.area == 4

    blue = Color(0, 0, 100)
    polyface.colors = (blue,)
    polyface.display_mode = 'Wireframe'
    assert polyface.colors == (blue,)
    assert polyface.display_mode == 'Wireframe'


def test_polyface3d_to_from_dict():
    """Test the to/from dict of Polyface3D objects."""
    grey = Color(100, 100, 100)
    polyface = DisplayPolyface3D(Polyface3D.from_box(2, 4, 2), [grey])
    polyface.display_mode = 'Wireframe'
    polyface_dict = polyface.to_dict()
    new_polyface = DisplayPolyface3D.from_dict(polyface_dict)
    assert isinstance(new_polyface, DisplayPolyface3D)
    assert new_polyface.to_dict() == polyface_dict

    assert len(new_polyface.vertices) == 8
    assert len(new_polyface.face_indices) == 6
    assert len(new_polyface.faces) == 6
    assert len(new_polyface.edge_indices) == 12
    assert len(new_polyface.edges) == 12
    assert len(new_polyface.naked_edges) == 0
    assert len(new_polyface.non_manifold_edges) == 0
    assert len(new_polyface.internal_edges) == 12
    assert new_polyface.area == 40
    assert new_polyface.volume == 16
    assert new_polyface.is_solid
    assert new_polyface.colors == (grey,)
    assert new_polyface.display_mode == 'Wireframe'



# coding=utf-8
from ladybug_geometry.geometry2d.pointvector import Point2D
from ladybug_geometry.geometry2d.polygon import Polygon2D
from ladybug.color import Color
from ladybug_display.geometry2d.polygon import DisplayPolygon2D


def test_display_p_gon2d_init():
    """Test the initialization of DisplayFace2D objects and basic properties."""
    grey = Color(100, 100, 100)
    pts = (Point2D(0, 0), Point2D(0, 2), Point2D(2, 2), Point2D(2, 0))
    p_gon = DisplayPolygon2D(Polygon2D(pts), grey)
    str(p_gon)  # test the string representation of the p_gon

    assert p_gon.color == grey
    assert p_gon.display_mode == 'Surface'
    assert isinstance(p_gon.vertices, tuple)
    assert len(p_gon.vertices) == 4
    for point in p_gon.vertices:
        assert isinstance(point, Point2D)
    assert p_gon.area == 4
    assert p_gon.perimeter == 8

    blue = Color(0, 0, 100)
    p_gon.color = blue
    p_gon.display_mode = 'Wireframe'
    assert p_gon.color == blue
    assert p_gon.display_mode == 'Wireframe'


def test_p_gon2d_to_from_dict():
    """Test the to/from dict of Face2D objects."""
    grey = Color(100, 100, 100)
    pts = (Point2D(0, 0), Point2D(0, 2), Point2D(2, 2), Point2D(2, 0))
    p_gon = DisplayPolygon2D(Polygon2D(pts), grey)
    p_gon.display_mode = 'Wireframe'
    p_gon_dict = p_gon.to_dict()
    new_p_gon = DisplayPolygon2D.from_dict(p_gon_dict)
    assert isinstance(new_p_gon, DisplayPolygon2D)
    assert new_p_gon.to_dict() == p_gon_dict



# coding=utf-8
import pytest

from ladybug_geometry.geometry2d.arc import Arc2D
from ladybug_geometry.geometry2d.pointvector import Point2D
from ladybug.color import Color
from ladybug_display.geometry2d.arc import DisplayArc2D
from ladybug_display.altnumber import default

import math


def test_display_arc2d_init():
    """Test the initialization of Arc2D objects and basic properties."""
    grey = Color(100, 100, 100)
    pt = Point2D(2, 0)
    arc = DisplayArc2D(Arc2D(pt, 1, 0, math.pi), grey)
    str(arc)  # test the string representation of the arc

    assert arc.color == grey
    assert arc.line_width == default
    assert arc.line_type == 'Continuous'
    assert arc.c == pt
    assert arc.r == 1
    assert arc.p1 == Point2D(3, 0)
    assert arc.p2.x == pytest.approx(1, rel=1e-3)
    assert arc.p2.y == pytest.approx(0, rel=1e-3)
    assert arc.length == pytest.approx(math.pi, rel=1e-3)
    assert arc.is_circle is False

    blue = Color(0, 0, 100)
    arc.color = blue
    arc.line_width = 2
    arc.line_type = 'Dashed'
    assert arc.color == blue
    assert arc.line_width == 2
    assert arc.line_type == 'Dashed'


def test_arc3_to_from_dict():
    """Test the initialization of Arc2D objects and basic properties."""
    grey = Color(100, 100, 100)
    pt = Point2D(2, 0)
    arc = DisplayArc2D(Arc2D(pt, 1, 0, math.pi), grey)
    arc_dict = arc.to_dict()
    new_arc = DisplayArc2D.from_dict(arc_dict)
    assert isinstance(new_arc, DisplayArc2D)
    assert new_arc.to_dict() == arc_dict



"""Geometry objects for display in 2D space."""

from .vector import DisplayVector2D
from .point import DisplayPoint2D
from .ray import DisplayRay2D
from .line import DisplayLineSegment2D
from .arc import DisplayArc2D
from .polyline import DisplayPolyline2D
from .polygon import DisplayPolygon2D
from .mesh import DisplayMesh2D



# coding=utf-8
"""Utilities to convert any Ladybug Display dictionary to Python objects."""

from ladybug_display.geometry2d import DisplayVector2D, DisplayPoint2D, \
    DisplayRay2D, DisplayLineSegment2D, DisplayPolyline2D, DisplayArc2D, \
    DisplayPolygon2D, DisplayMesh2D
from ladybug_display.geometry3d import DisplayVector3D, DisplayPoint3D, \
    DisplayRay3D, DisplayPlane, DisplayLineSegment3D, DisplayPolyline3D, DisplayArc3D, \
    DisplayFace3D, DisplayMesh3D, DisplayPolyface3D, DisplaySphere, DisplayCone, \
    DisplayCylinder


def dict_to_object(display_dict, raise_exception=True):
    """
    Args:
        display_dict (dict): A dictionary of any Ladybug Display geometry object.
        raise_exception (bool): Boolean to note whether an exception should be
            raised if the object is not identified as a part of
            ladybug_display. (Default: True).

    Returns:
        A Python object derived from the input display_dict.
    """
    lbt_types = {
        'DisplayVector2D': DisplayVector2D,
        'DisplayPoint2D': DisplayPoint2D,
        'DisplayRay2D': DisplayRay2D,
        'DisplayLineSegment2D': DisplayLineSegment2D,
        'DisplayArc2D': DisplayArc2D,
        'DisplayPolyline2D': DisplayPolyline2D,
        'DisplayPolygon2D': DisplayPolygon2D,
        'DisplayMesh2D': DisplayMesh2D,
        'DisplayVector3D': DisplayVector3D,
        'DisplayPoint3D': DisplayPoint3D,
        'DisplayRay3D': DisplayRay3D,
        'DisplayLineSegment3D': DisplayLineSegment3D,
        'DisplayArc3D': DisplayArc3D,
        'DisplayPolyline3D': DisplayPolyline3D,
        'DisplayMesh3D': DisplayMesh3D,
        'DisplayPlane': DisplayPlane,
        'DisplayPolyface3D': DisplayPolyface3D,
        'DisplayFace3D': DisplayFace3D,
        'DisplaySphere': DisplaySphere,
        'DisplayCone': DisplayCone,
        'DisplayCylinder': DisplayCylinder
    }

    # Get the ladybug_geometry object 'Type'
    try:
        obj_type = display_dict['type']
    except KeyError:
        raise ValueError('Ladybug dictionary lacks required "type" key.')

    # Build a new Ladybug Python Object based on the "Type"
    try:
        lbt_class = lbt_types[obj_type]
        return lbt_class.from_dict(display_dict)
    except KeyError:
        if raise_exception:
            raise ValueError(
                '{} is not a recognized ladybug display type'.format(obj_type))
        else:
            return None



# coding=utf-8
from ladybug_geometry.geometry3d.pointvector import Point3D, Vector3D
from ladybug_geometry.geometry3d.line import LineSegment3D
from ladybug.color import Color
from ladybug_display.geometry3d.line import DisplayLineSegment3D
from ladybug_display.altnumber import default


def test_display_linesegment3d_init():
    """Test the initialization of LineSegment3D objects and basic properties."""
    grey = Color(100, 100, 100)
    pt = Point3D(2, 0, 2)
    vec = Vector3D(0, 2, 0)
    seg = DisplayLineSegment3D(LineSegment3D(pt, vec), grey)
    str(seg)  # test the string representation of the line segment

    assert seg.color == grey
    assert seg.line_width == default
    assert seg.line_type == 'Continuous'
    assert seg.p == Point3D(2, 0, 2)
    assert seg.v == Vector3D(0, 2, 0)
    assert seg.p1 == Point3D(2, 0, 2)
    assert seg.p2 == Point3D(2, 2, 2)
    assert seg.length == 2

    blue = Color(0, 0, 100)
    seg.color = blue
    seg.line_width = 2
    seg.line_type = 'Dashed'
    assert seg.color == blue
    assert seg.line_width == 2
    assert seg.line_type == 'Dashed'


def test_linesegment3_to_from_dict():
    """Test the to/from dict of DisplayLineSegment3D objects."""
    grey = Color(100, 100, 100)
    pt = Point3D(2, 0, 2)
    vec = Vector3D(0, 2, 0)
    seg = DisplayLineSegment3D(LineSegment3D(pt, vec), grey)
    seg.line_type = 'Dashed'
    seg_dict = seg.to_dict()
    new_seg = DisplayLineSegment3D.from_dict(seg_dict)
    assert isinstance(new_seg, DisplayLineSegment3D)
    assert new_seg.to_dict() == seg_dict



"""A point that can be displayed in 3D space."""
from ladybug_geometry.geometry3d.pointvector import Point3D
from ladybug.color import Color

from ._base import _SingleColorBase3D
from ladybug_display.altnumber import default
from ladybug_display.typing import float_positive


class DisplayPoint3D(_SingleColorBase3D):
    """A point in 3D space with display properties.

    Args:
        geometry: A ladybug-geometry Point3D object.
        color: A ladybug Color object. If None, a default black color will be
            used. (Default: None).
        radius: Number for radius in pixels (for the screen) or millimeters
            (in print). This can also be the Default object to indicate that the
            default settings of the interface should be used.

    Properties:
        * geometry
        * color
        * radius
        * x
        * y
        * z
        * user_data
    """
    __slots__ = ('_radius',)

    def __init__(self, geometry, color=None, radius=default):
        """Initialize base with shade object."""
        assert isinstance(geometry, Point3D), '\
            Expected ladybug_geometry Point3D. Got {}'.format(type(geometry))
        _SingleColorBase3D.__init__(self, geometry, color)
        self.radius = radius

    @classmethod
    def from_dict(cls, data):
        """Initialize a DisplayPoint3D from a dictionary.

        Args:
            data: A dictionary representation of an DisplayPoint3D object.
        """
        assert data['type'] == 'DisplayPoint3D', \
            'Expected DisplayPoint3D dictionary. Got {}.'.format(data['type'])
        color = Color.from_dict(data['color']) if 'color' in data and data['color'] \
            is not None else None
        rad = default if 'radius' not in data or \
            data['radius'] == default.to_dict() else data['radius']
        geo = cls(Point3D.from_dict(data['geometry']), color)
        if 'user_data' in data and data['user_data'] is not None:
            geo.user_data = data['user_data']
        return geo

    @property
    def radius(self):
        """Get or set a number for the radius of this object."""
        return self._radius

    @radius.setter
    def radius(self, value):
        if value == default:
            self._radius = default
        else:
            self._radius = float_positive(value, 'point radius')

    @property
    def x(self):
        """Get the X coordinate."""
        return self._geometry.x

    @property
    def y(self):
        """Get the Y coordinate."""
        return self._geometry.y

    @property
    def z(self):
        """Get the Z coordinate."""
        return self._geometry.z

    def distance_to_point(self, point):
        """Get the distance from this point to another DisplayPoint3D."""
        return self.geometry.distance_to_point(point.geometry)

    def to_dict(self):
        """Return DisplayPoint3D as a dictionary."""
        base = {'type': 'DisplayPoint3D'}
        base['geometry'] = self._geometry.to_dict()
        base['color'] = self.color.to_dict()
        base['radius'] = default.to_dict() if self.radius == default else self.radius
        if self.user_data is not None:
            base['user_data'] = self.user_data
        return base

    def __copy__(self):
        new_g = DisplayPoint3D(self.geometry, self.color, self.radius)
        new_g._user_data = None if self.user_data is None else self.user_data.copy()
        return new_g

    def __getitem__(self, key):
        return (self.x, self.y, self.z)[key]

    def __iter__(self):
        return iter((self.x, self.y, self.z))

    def __repr__(self):
        return 'DisplayPoint3D: {}'.format(self.geometry)



"""A polyline that can be displayed in 3D space."""
from ladybug_geometry.geometry3d.polyline import Polyline3D
from ladybug.color import Color

from ladybug_display.altnumber import default
from ._base import _LineCurveBase3D


class DisplayPolyline3D(_LineCurveBase3D):
    """A polyline in 3D space with display properties.

    Args:
        geometry: A ladybug-geometry Polyline3D object.
        color: A ladybug Color object. If None, a default black color will be
            used. (Default: None).
        line_width: Number for line width in pixels (for the screen) or  millimeters
            (in print). This can also be the Default object to indicate that the
            default settings of the interface should be used.
        line_type: Get or set text to indicate the type of line to display. 
            Choose from the following. (Default: "Continuous")

            * Continuous
            * Dashed
            * Dotted
            * DashDot
            * Center
            * Border
            * Hidden

    Properties:
        * geometry
        * color
        * line_width
        * line_type
        * vertices
        * segments
        * interpolated
        * p1
        * p2
        * length
        * min
        * max
        * user_data
    """
    __slots__ = ()

    def __init__(self, geometry, color=None, line_width=default, line_type='Continuous'):
        """Initialize base with shade object."""
        assert isinstance(geometry, Polyline3D), '\
            Expected ladybug_geometry Polyline3D. Got {}'.format(type(geometry))
        _LineCurveBase3D.__init__(self, geometry, color, line_width, line_type)

    @classmethod
    def from_dict(cls, data):
        """Initialize a DisplayPolyline3D from a dictionary.

        Args:
            data: A dictionary representation of an DisplayPolyline3D object.
        """
        assert data['type'] == 'DisplayPolyline3D', \
            'Expected DisplayPolyline3D dictionary. Got {}.'.format(data['type'])
        color = Color.from_dict(data['color']) if 'color' in data and data['color'] \
            is not None else None
        lw = default if 'line_width' not in data or \
            data['line_width'] == default.to_dict() else data['line_width']
        lt = data['line_type'] if 'line_type' in data and data['line_type'] \
            is not None else 'Continuous'
        geo = cls(Polyline3D.from_dict(data['geometry']), color, lw, lt)
        if 'user_data' in data and data['user_data'] is not None:
            geo.user_data = data['user_data']
        return geo

    @property
    def vertices(self):
        """Get a tuple of Point3D for the vertices of the polyline."""
        return self._geometry.vertices

    @property
    def segments(self):
        """Get a tuple of LineSegment3D for the segments of the polyline."""
        return self._geometry.segments

    @property
    def interpolated(self):
        """Get a boolean for whether the polyline should be interpreted as interpolated.
        """
        return self._geometry.interpolated

    @property
    def p1(self):
        """Get a Point3D representing the first end point of the polyline."""
        return self._geometry.p1

    @property
    def p2(self):
        """Get a Point3D representing the second end point of the polyline."""
        return self._geometry.p2

    @property
    def length(self):
        """Get a number for the length of the polyline."""
        return self._geometry.length

    @property
    def min(self):
        """Get a Point3D for the minimum of the bounding box around the object."""
        return self._geometry.min

    @property
    def max(self):
        """Get a Point3D for the maximum of the bounding box around the object."""
        return self._geometry.max

    def to_dict(self):
        """Return DisplayPolyline3D as a dictionary."""
        base = {'type': 'DisplayPolyline3D'}
        base['geometry'] = self._geometry.to_dict()
        base['color'] = self.color.to_dict()
        base['line_width'] = default.to_dict() if \
            self.line_width == default else self.line_width
        base['line_type'] = self.line_type
        if self.user_data is not None:
            base['user_data'] = self.user_data
        return base

    def __copy__(self):
        new_g = DisplayPolyline3D(
            self.geometry, self.color, self.line_width, self.line_type)
        new_g._user_data = None if self.user_data is None else self.user_data.copy()
        return new_g

    def __repr__(self):
        return 'DisplayPolyline3D: {}'.format(self.geometry)



# coding=utf-8
from ladybug_geometry.geometry2d.pointvector import Point2D
from ladybug_geometry.geometry2d.line import LineSegment2D
from ladybug_geometry.geometry2d.polyline import Polyline2D
from ladybug.color import Color
from ladybug_display.geometry2d.polyline import DisplayPolyline2D
from ladybug_display.altnumber import default


def test_display_polyline2d_init():
    """Test the initialization of DisplayPolyline2D objects and basic properties."""
    grey = Color(100, 100, 100)
    pts = (Point2D(0, 0), Point2D(2, 0), Point2D(2, 2), Point2D(0, 2))
    pline = DisplayPolyline2D(Polyline2D(pts), grey)
    str(pline)  # test the string representation of the polyline

    assert pline.color == grey
    assert pline.line_width == default
    assert pline.line_type == 'Continuous'

    assert isinstance(pline.vertices, tuple)
    assert len(pline.vertices) == 4
    for point in pline.vertices:
        assert isinstance(point, Point2D)

    assert isinstance(pline.segments, tuple)
    assert len(pline.segments) == 3
    for seg in pline.segments:
        assert isinstance(seg, LineSegment2D)
        assert seg.length == 2

    assert pline.p1 == pts[0]
    assert pline.p2 == pts[-1]
    assert pline.length == 6

    blue = Color(0, 0, 100)
    pline.color = blue
    pline.line_width = 2
    pline.line_type = 'Dashed'
    assert pline.color == blue
    assert pline.line_width == 2
    assert pline.line_type == 'Dashed'


def test_polyline2d_to_from_dict():
    """Test the to/from dict of DisplayPolyline2D objects."""
    grey = Color(100, 100, 100)
    pts = (Point2D(0, 0), Point2D(2, 0), Point2D(2, 2), Point2D(0, 2))
    pline = DisplayPolyline2D(Polyline2D(pts), grey)
    pline.line_width = 2
    pline.line_type = 'Dashed'
    pline_dict = pline.to_dict()
    new_pline = DisplayPolyline2D.from_dict(pline_dict)
    assert isinstance(new_pline, DisplayPolyline2D)
    assert new_pline.to_dict() == pline_dict



# coding=utf-8
from ladybug_geometry.geometry3d.pointvector import Point3D
from ladybug_geometry.geometry3d.sphere import Sphere
from ladybug.color import Color
from ladybug_display.geometry3d.sphere import DisplaySphere


def test_display_sphere_init():
    """Test the initialization of Sphere objects and basic properties."""
    grey = Color(100, 100, 100)
    pt = Point3D(2, 0, 2)
    r = 3
    sp = DisplaySphere(Sphere(pt, r), grey)
    str(sp)  # test the string representation of the line segment

    assert sp.color == grey
    assert sp.display_mode == 'Surface'
    assert sp.center == Point3D(2, 0, 2)
    assert sp.radius == 3
    assert sp.min.z == -1
    assert sp.max.z == 5
    assert sp.diameter == 6
    assert isinstance(sp.circumference, float)
    assert isinstance(sp.area, float)
    assert isinstance(sp.volume, float)

    blue = Color(0, 0, 100)
    sp.color = blue
    sp.display_mode = 'Wireframe'
    assert sp.color == blue
    assert sp.display_mode == 'Wireframe'


def test_sphere_to_from_dict():
    """Test the Sphere to_dict and from_dict methods."""
    grey = Color(100, 100, 100)
    pt = Point3D(2, 0, 2)
    r = 3
    sp = DisplaySphere(Sphere(pt, r), grey)

    sp_d = sp.to_dict()
    new_sp = DisplaySphere.from_dict(sp_d)
    assert isinstance(new_sp, DisplaySphere)
    assert new_sp.to_dict() == sp_d



"""A polyface in 3D space with display properties."""
from ladybug_geometry.geometry3d.polyface import Polyface3D
from ladybug.color import Color

from .._base import DISPLAY_MODES
from ._base import _DisplayBase3D


class DisplayPolyface3D(_DisplayBase3D):
    """A polyface in 3D space with display properties.

    Args:
        geometry: A ladybug-geometry Polyface3D.
        colors: A list of colors that correspond to either the faces of the polyface
            or the vertices of the polyface. It can also be a single color for the
            entire polyface. (Default: None).
        display_mode: Text to indicate the display mode (surface, wireframe, etc.).
            Choose from the following. (Default: Surface).

            * Surface
            * SurfaceWithEdges
            * Wireframe
            * Points

    Properties:
        * geometry
        * colors
        * display_mode
        * vertices
        * faces
        * edges
        * naked_edges
        * internal_edges
        * non_manifold_edges
        * face_indices
        * edge_indices
        * edge_types
        * min
        * max
        * center
        * area
        * volume
        * is_solid
        * user_data
    """
    __slots__ = ('_colors', '_display_mode')

    def __init__(self, geometry, colors=None, display_mode='Surface'):
        """Initialize base with shade object."""
        assert isinstance(geometry, Polyface3D), '\
            Expected ladybug_geometry Polyface3D. Got {}'.format(type(geometry))
        _DisplayBase3D.__init__(self, geometry)
        self.colors = colors
        self.display_mode = display_mode

    @classmethod
    def from_dict(cls, data):
        """Initialize a DisplayPolyface3D from a dictionary.

        Args:
            data: A dictionary representation of an DisplayPolyface3D object.
        """
        assert data['type'] == 'DisplayPolyface3D', \
            'Expected DisplayPolyface3D dictionary. Got {}.'.format(data['type'])
        colors = [Color.from_dict(c) for c in data['colors']] if 'colors' in data \
            and data['colors'] is not None else None
        d_mode = data['display_mode'] if 'display_mode' in data and \
            data['display_mode'] is not None else 'Surface'
        geo = cls(Polyface3D.from_dict(data['geometry']), colors, d_mode)
        if 'user_data' in data and data['user_data'] is not None:
            geo.user_data = data['user_data']
        return geo

    @property
    def colors(self):
        """Get or set a tuple of colors for this object.

        The length of this tuple must be either equal to the number or faces or
        vertices of the polyface. Or it can be a tuple with a single color for the
        whole geometry object.
        """
        return self._colors

    @colors.setter
    def colors(self, col):
        if col is None:
            col = (Color(0, 0, 0),)
        else:
            assert isinstance(col, (list, tuple)), \
                'colors should be a list or tuple. Got {}'.format(type(col))
            if isinstance(col, list):
                col = tuple(col)
            l_col = len(col)
            if l_col == 1 or l_col == len(self.faces) or l_col == len(self.vertices):
                pass
            elif l_col == 0:
                col = (Color(0, 0, 0),)
            else:
                msg = 'Number of colors ({}) does not match the number of faces ' \
                    '({}) nor the number of vertices ({}).'.format(
                        len(col), len(self.faces), len(self.vertices))
                raise ValueError(msg)
            assert all(isinstance(v, Color) for v in col), 'Expected Color for ' \
                'ladybug_display object color.'
        self._colors = col

    @property
    def display_mode(self):
        """Get or set text to indicate the display mode."""
        return self._display_mode

    @display_mode.setter
    def display_mode(self, value):
        clean_input = value.lower()
        for key in DISPLAY_MODES:
            if key.lower() == clean_input:
                value = key
                break
        else:
            raise ValueError(
                'display_mode {} is not recognized.\nChoose from the '
                'following:\n{}'.format(value, DISPLAY_MODES))
        self._display_mode = value

    @property
    def vertices(self):
        """Get a tuple of vertices in the polyface."""
        return self._geometry.vertices

    @property
    def faces(self):
        """Get a tuple of all Face3D objects making up this polyface."""
        return self._geometry.faces

    @property
    def edges(self):
        """Get a tuple of all edges in this polyface as LineSegment3D objects."""
        return self._geometry.edges

    @property
    def naked_edges(self):
        """Get a tuple of all naked edges in this polyface as LineSegment3D objects."""
        return self._geometry.naked_edges

    @property
    def internal_edges(self):
        """Get a tuple of all internal edges in this polyface as LineSegment3D objects.
        """
        return self._geometry.internal_edges

    @property
    def non_manifold_edges(self):
        """Get a tuple of all non-manifold edges as LineSegment3D objects."""
        return self._geometry.non_manifold_edges

    @property
    def face_indices(self):
        """Get a tuple of face tuples with integers corresponding to indices of vertices.
        """
        return self._geometry.face_indices

    @property
    def edge_indices(self):
        """Get a tuple of edge tuples with integers corresponding to indices of vertices.
        """
        return self._geometry.edge_indices

    @property
    def edge_types(self):
        """Get a tuple of integers for each edge that denotes the type of edge.

        0 denotes a naked edge, 1 denotes an internal edge, and anything higher is a
        non-manifold edge.
        """
        return self._geometry.edge_types

    @property
    def min(self):
        """Get a Point3D for the minimum of the bounding box around the object."""
        return self._geometry.min

    @property
    def max(self):
        """Get a Point3D for the maximum of the bounding box around the object."""
        return self._geometry.max

    @property
    def center(self):
        """Get a Point3D for the center of the bounding box around the object."""
        return self._geometry.center

    @property
    def area(self):
        """Get the area of the polyface."""
        return self._geometry.area

    @property
    def volume(self):
        """Get the volume of the polyface."""
        return self._geometry.volume

    @property
    def is_solid(self):
        """Get a boolean to note whether the polyface is solid (True) or is open (False).
        """
        return self._geometry.is_solid

    def to_dict(self):
        """Return DisplayPolyface3D as a dictionary."""
        base = {'type': 'DisplayPolyface3D'}
        base['geometry'] = self._geometry.to_dict()
        base['colors'] = [c.to_dict() for c in self.colors]
        base['display_mode'] = self.display_mode
        if self.user_data is not None:
            base['user_data'] = self.user_data
        return base

    def __copy__(self):
        new_g = DisplayPolyface3D(self.geometry, self.colors, self.display_mode)
        new_g._user_data = None if self.user_data is None else self.user_data.copy()
        return new_g

    def __repr__(self):
        return 'DisplayPolyface3D: {}'.format(self.geometry)



# coding=utf-8
from ladybug_geometry.geometry3d.pointvector import Point3D, Vector3D
from ladybug_geometry.geometry3d.plane import Plane

from ladybug.color import Color
from ladybug_display.geometry3d.plane import DisplayPlane


def test_plane_init():
    """Test the initialization of Plane objects and basic properties."""
    grey = Color(100, 100, 100)
    pt = Point3D(2, 0, 2)
    vec = Vector3D(0, 2, 0)
    plane = DisplayPlane(Plane(vec, pt), grey)
    str(plane)  # test the string representation

    assert not plane.show_axes
    assert not plane.show_grid
    assert plane.o == Point3D(2, 0, 2)
    assert plane.n == Vector3D(0, 1, 0)
    assert plane.x == Vector3D(1, 0, 0)
    assert plane.y == Vector3D(0, 0, -1)
    assert plane.k == 0

    blue = Color(0, 0, 100)
    plane.color = blue
    plane.show_axes = True
    plane.show_grid = True
    assert plane.color == blue
    assert plane.show_axes
    assert plane.show_grid


def test_plane_to_from_dict():
    """Test the initialization of Plane objects and basic properties."""
    grey = Color(100, 100, 100)
    pt = Point3D(2, 0, 2)
    vec = Vector3D(0, 2, 0)
    plane = DisplayPlane(Plane(vec, pt), grey)
    plane.show_axes = True
    plane.show_grid = True
    plane_dict = plane.to_dict()
    new_plane = DisplayPlane.from_dict(plane_dict)
    assert isinstance(new_plane, DisplayPlane)
    assert new_plane.to_dict() == plane_dict



# coding=utf-8
from __future__ import division
import collections

from ladybug.legend import Legend, LegendParameters, LegendParametersCategorized
from ladybug.graphic import GraphicContainer
from ladybug.datatype.base import DataTypeBase

from ladybug_geometry.geometry2d import Vector2D, Point2D, Ray2D, LineSegment2D, \
    Polyline2D, Arc2D, Polygon2D, Mesh2D
from ladybug_geometry.geometry3d import Vector3D, Point3D, Ray3D, Plane, LineSegment3D, \
    Polyline3D, Arc3D, Face3D, Mesh3D, Polyface3D, Sphere, Cone, Cylinder
from .geometry2d import DisplayVector2D, DisplayPoint2D, \
    DisplayRay2D, DisplayLineSegment2D, DisplayPolyline2D, DisplayArc2D, \
    DisplayPolygon2D, DisplayMesh2D
from .geometry3d import DisplayVector3D, DisplayPoint3D, \
    DisplayRay3D, DisplayPlane, DisplayLineSegment3D, DisplayPolyline3D, DisplayArc3D, \
    DisplayFace3D, DisplayMesh3D, DisplayPolyface3D, DisplaySphere, DisplayCone, \
    DisplayCylinder
from ladybug_geometry.bounding import bounding_box
from ladybug_geometry.dictutil import geometry_dict_to_object

from ._base import DISPLAY_MODES
from .typing import int_in_range, valid_string
from .dictutil import dict_to_object

GEOMETRY_UNION = (
    Vector2D, Point2D, Ray2D, LineSegment2D, Polyline2D, Arc2D, Polygon2D,
    Mesh2D, Vector3D, Point3D, Ray3D, Plane, LineSegment3D,
    Polyline3D, Arc3D, Face3D, Mesh3D, Polyface3D, Sphere, Cone, Cylinder
)
DISPLAY_UNION = (
    DisplayVector2D, DisplayPoint2D, DisplayRay2D, DisplayLineSegment2D,
    DisplayPolyline2D, DisplayArc2D, DisplayPolygon2D, DisplayMesh2D,
    DisplayVector3D, DisplayPoint3D, DisplayRay3D, DisplayPlane, DisplayLineSegment3D,
    DisplayPolyline3D, DisplayArc3D, DisplayFace3D, DisplayMesh3D,
    DisplayPolyface3D, DisplaySphere, DisplayCone, DisplayCylinder
)


class _VisualizationBase(object):
    """A base class for visualization objects.

    Args:
        identifier: Text string for a unique object ID. Must be less than 100
            characters and not contain spaces or special characters.

    Properties:
        * identifier
        * display_name
        * full_id
        * user_data
    """
    __slots__ = ('_identifier', '_display_name', '_user_data')

    def __init__(self, identifier):
        """Initialize base object."""
        self.identifier = identifier
        self._display_name = None
        self._user_data = None

    @property
    def identifier(self):
        """Get or set a text string for the unique object identifier.

        This identifier remains constant as the object is mutated, copied, and
        serialized to different formats.
        """
        return self._identifier

    @identifier.setter
    def identifier(self, value):
        self._identifier = valid_string(value, 'visualization object identifier')

    @property
    def display_name(self):
        """Get or set text for the object name without any character restrictions.

        This is typically used to set the layer of the object in the interface that
        renders the VisualizationSet. A :: in the display_name can be used to denote
        sub-layers following a convention of ParentLayer::SubLayer.

        If not set, the display_name will be equal to the object identifier.
        """
        if self._display_name is None:
            return self._identifier
        return self._display_name

    @display_name.setter
    def display_name(self, value):
        if value is not None:
            try:
                value = str(value)
            except UnicodeEncodeError:  # Python 2 machine lacking the character set
                pass  # keep it as unicode
        self._display_name = value

    @property
    def full_id(self):
        """Get a string with both the object display_name and identifier.

        This is formatted as display_name[identifier].

        This is useful in error messages to give users an easy means of finding
        invalid objects within models. If there is no display_name assigned,
        only the identifier will be returned.
        """
        if self._display_name is None:
            return self._identifier
        else:
            return '{}[{}]'.format(self._display_name, self._identifier)

    @property
    def user_data(self):
        """Get or set an optional dictionary for additional meta data for this object.

        This will be None until it has been set. All keys and values of this
        dictionary should be of a standard Python type to ensure correct
        serialization of the object to/from JSON (eg. str, float, int, list, dict)
        """
        return self._user_data

    @user_data.setter
    def user_data(self, value):
        if value is not None:
            assert isinstance(value, dict), 'Expected dictionary for visualization ' \
                'object user_data. Got {}.'.format(type(value))
        self._user_data = value


class VisualizationSet(_VisualizationBase):
    """A visualization set containing analysis and context geometry to be visualized.

    Args:
        identifier: Text string for a unique object ID. Must be less than 100
            characters and not contain spaces or special characters.
        geometry: A list of AnalysisGeometry and ContextGeometry objects to display
            in the visualization. Each geometry object will typically be translated
            to its own layer within the interface that renders the VisualizationSet.

    Properties:
        * identifier
        * display_name
        * geometry
        * min_point
        * max_point
        * user_data
    """
    __slots__ = ('_geometry', '_min_point', '_max_point')

    def __init__(self, identifier, geometry):
        """Initialize VisualizationSet."""
        _VisualizationBase.__init__(self, identifier)  # process the identifier
        self.geometry = geometry
        self._min_point = None
        self._max_point = None

    @classmethod
    def from_dict(cls, data):
        """Create an VisualizationSet from a dictionary.

        Args:
            data: A python dictionary in the following format

        .. code-block:: python

            {
            "type": "VisualizationSet",
            "identifier": "",  # unique object identifier
            "geometry": []  # list of AnalysisGeometry and ContextGeometry objects
            }
        """
        # check the type key
        assert data['type'] == 'VisualizationSet', \
            'Expected VisualizationSet, Got {}.'.format(data['type'])
        # re-serialize the context and analysis geometry
        geos = []
        for geo_data in data['geometry']:
            if geo_data['type'] == 'AnalysisGeometry':
                geos.append(AnalysisGeometry.from_dict(geo_data))
            else:
                geos.append(ContextGeometry.from_dict(geo_data))
        new_obj = cls(data['identifier'], geos)
        if 'display_name' in data and data['display_name'] is not None:
            new_obj.display_name = data['display_name']
        if 'user_data' in data and data['user_data'] is not None:
            new_obj.user_data = data['user_data']
        return new_obj

    @classmethod
    def from_single_analysis_geo(
            cls, identifier, geometry, values, legend_parameters=None,
            data_type=None, unit=None):
        """Create an VisualizationSet from a raw geometry object and a list of values.

        Args:
            identifier: Text string for a unique object ID. Must be less than 100
                characters and not contain spaces or special characters.
            geometry: A list of ladybug-geometry objects that is aligned with the
                values. The length of this list should usually be equal to the total
                number of values in each data_set, indicating that each geometry gets
                a single color. Alternatively, if all of the geometry objects are
                meshes, the number of values in the data can be equal to the total
                number of faces across the meshes or the total number of vertices
                across the meshes.
            values: A list of numerical values that will be used to generate the
                visualization colors.
            legend_parameters: An Optional LegendParameters object to override default
                parameters of the legend. None indicates that default legend parameters
                will be used. (Default: None).
            data_type: Optional DataType from the ladybug datatype subpackage (ie.
                Temperature()) , which will be used to assign default legend properties.
                If None, the legend associated with this object will contain no units
                unless a unit below is specified. (Default: None).
            unit: Optional text string for the units of the values. (ie. "C"). If None
                or empty, the default units of the data_type will be used. If no data
                type is specified in this case, this will simply be an empty
                string. (Default: None).
        """
        viz_data = VisualizationData(values, legend_parameters, data_type, unit)
        a_geo = AnalysisGeometry('{}_Geometry'.format(identifier), geometry, [viz_data])
        return cls(identifier, (a_geo,))

    @property
    def geometry(self):
        """Get or set a tuple of AnalysisGeometry and ContextGeometry objects."""
        return self._geometry

    @geometry.setter
    def geometry(self, value):
        assert isinstance(value, (list, tuple)), 'Expected list or tuple for' \
            ' VisualizationSet geometry. Got {}.'.format(type(value))
        if not isinstance(value, tuple):
            value = tuple(value)
        for geo in value:
            self._check_geometry(geo)
        self._geometry = value

    @property
    def min_point(self):
        """A Point3D for the minimum bounding box vertex around all of the geometry."""
        if self._min_point is None:
            self._calculate_min_max()
        return self._min_point

    @property
    def max_point(self):
        """A Point3D for the maximum bounding box vertex around all of the geometry."""
        if self._max_point is None:
            self._calculate_min_max()
        return self._max_point

    def add_geometry(self, geometry, insert_index=None):
        """Add a data set to this AnalysisGeometry object.

        Args:
            data: A VisualizationData object to be added to this AnalysisGeometry.
            insert_index: An integer for the index at which the data should be
                inserted. If None, the data will be appended to the end. (Default: None).
        """
        self._check_geometry(geometry)
        if insert_index is None:
            self._geometry = self._geometry + (geometry,)
        else:
            geos_list = list(self._geometry)
            geos_list.insert(insert_index, geometry)
            self._geometry = tuple(geos_list)

    def remove_geometry(self, geo_index):
        """Remove a geometry object from this VisualizationSet.

        Args:
            geo_index: An integer for the geometry index to be removed.
        """
        assert geo_index < len(self._geometry), 'geo_index ({}) must be less than ' \
            'the number of items in the data_sets ({}).'.format(
                geo_index, len(self._geometry))
        geos_list = list(self._geometry)
        geos_list.pop(geo_index)
        self._geometry = tuple(geos_list)

    def check_duplicate_identifiers(self, raise_exception=True, detailed=False):
        """Check that there are no duplicate geometry object identifiers in the set.

        Args:
            raise_exception: Boolean to note whether a ValueError should be raised
                if duplicate identifiers are found. (Default: True).
            detailed: Boolean for whether the returned object is a detailed list of
                dicts with error info or a string with a message. (Default: False).

        Returns:
            A string with the message or a list with a dictionary if detailed is True.
        """
        detailed = False if raise_exception else detailed
        obj_id_iter = (obj.identifier for obj in self.geometry)
        dup = [t for t, c in collections.Counter(obj_id_iter).items() if c > 1]
        if len(dup) != 0:
            if detailed:
                err_list = []
                for dup_id in dup:
                    msg = 'There is a duplicated geometry identifier: {}'.format(dup_id)
                    dup_dict = {
                        'type': 'ValidationError',
                        'element_type': 'Geometry',
                        'element_id': dup_id,
                        'element_name': dup_id,
                        'message': msg
                    }
                    err_list.append(dup_dict)
                return err_list
            msg = 'The following duplicated Geometry identifiers were found:\n{}'.format(
                '\n'.join(dup))
            if raise_exception:
                raise ValueError(msg)
            return msg
        return [] if detailed else ''

    def graphic_container(self, geo_index=0, data_index=None,
                          min_point=None, max_point=None):
        """Get a Ladybug GraphicContainer object, which can be used to draw legends.

        Args:
            geo_index: Integer for the index of the geometry for which a
                GraphicContainer will be returned. Note that this index must refer
                to an analysis geometry in order to produce a valid graphic
                container. (Default: 0).
            data_index: Integer for the index of the data set for which a
                GraphicContainer will be returned. If None, the active_data set
                will be used. (Default: None).
            min_point: An optional Point3D to denote the minimum bounding box
                for the graphic container. If None, this object's own min_point
                will be used, which corresponds to the bounding box around
                the geometry. (Default: None).
            max_point: An optional Point3D to denote the maximum bounding box
                for the graphic container. If None, this object's own max_point
                will be used, which corresponds to the bounding box around
                the geometry. (Default: None).
        """
        # check to be sure that there is analysis geometry
        geo_obj = self.geometry[geo_index]
        assert isinstance(geo_obj, AnalysisGeometry), 'VisualizationSet geo_index ' \
            'must refer to an  AnalysisGeometry in order to use ' \
            'graphic_container method.'
        # ensure that min and max points always make sense
        min_point = self.min_point if min_point is None else min_point
        max_point = self.max_point if max_point is None else max_point
        # return the Graphic Container for the correct data set
        data_index = geo_obj.active_data if data_index is None else data_index
        dat_set = geo_obj.data_sets[data_index]
        return dat_set.graphic_container(min_point, max_point)

    def to_dict(self):
        """Get VisualizationSet as a dictionary."""
        base = {
            'type': 'VisualizationSet',
            'identifier': self.identifier,
            'geometry': [geo_obj.to_dict() for geo_obj in self.geometry]
        }
        if self._display_name is not None:
            base['display_name'] = self.display_name
        if self.user_data is not None:
            base['user_data'] = self.user_data
        return base

    def _check_geometry(self, geo):
        """Check that the geometry object is valid."""
        assert isinstance(geo, (AnalysisGeometry, ContextGeometry)), 'Expected ' \
            'AnalysisGeometry or ContextGeometry for VisualizationSet geometry. ' \
            'Got {}.'.format(type(geo))

    def _calculate_min_max(self):
        """Calculate maximum and minimum Point3D for this object."""
        all_geo = []
        for geo_obj in self.geometry:
            all_geo.append(geo_obj.min_point)
            all_geo.append(geo_obj.max_point)
        if len(all_geo) != 0:
            self._min_point, self._max_point = bounding_box(all_geo)

    def ToString(self):
        """Overwrite .NET ToString."""
        return self.__repr__()

    def __len__(self):
        """Return number of geometries on the object."""
        return len(self.geometry)

    def __getitem__(self, key):
        """Return one of the geometries."""
        return self.geometry[key]

    def __iter__(self):
        """Iterate through the geometries."""
        return iter(self.geometry)

    def __repr__(self):
        """VisualizationSet representation."""
        return 'Visualization Set: {}'.format(self.display_name)


class ContextGeometry(_VisualizationBase):
    """An object representing context geometry to display.

    Args:
        identifier: Text string for a unique object ID. Must be less than 100
            characters and not contain spaces or special characters.
        geometry: A list of ladybug-geometry or ladybug-display objects that gives
            context to analysis geometry or other aspects of the visualization.
            Typically, these will display in wireframe around the geometry, though
            the properties of display geometry can be used to customize the
            visualization.

    Properties:
        * identifier
        * display_name
        * geometry
        * min_point
        * max_point
        * user_data
    """
    __slots__ = ('_geometry', '_min_point', '_max_point')

    WIREFRAME_MAP = {
        Vector2D: (DisplayVector2D, None),
        Point2D: (DisplayPoint2D, None),
        Ray2D: (DisplayRay2D, None),
        LineSegment2D: (DisplayLineSegment2D, None),
        Polyline2D: (DisplayPolyline2D, None),
        Arc2D: (DisplayArc2D, None),
        Polygon2D: (DisplayPolygon2D, None, 'Wireframe'),
        Mesh2D: (DisplayMesh2D, None, 'Wireframe'),
        Vector3D: (DisplayVector3D, None),
        Point3D: (DisplayPoint3D, None),
        Ray3D: (DisplayRay3D, None),
        Plane: (DisplayPlane, None),
        LineSegment3D: (DisplayLineSegment3D, None),
        Polyline3D: (DisplayPolyline3D, None),
        Arc3D: (DisplayArc3D, None),
        Face3D: (DisplayFace3D, None, 'Wireframe'),
        Mesh3D: (DisplayMesh3D, None, 'Wireframe'),
        Polyface3D: (DisplayPolyface3D, None, 'Wireframe'),
        Sphere: (DisplaySphere, None, 'Wireframe'),
        Cone: (DisplayCone, None, 'Wireframe'),
        Cylinder: (DisplayCylinder, None, 'Wireframe')
    }

    def __init__(self, identifier, geometry):
        """Initialize ContextGeometry."""
        _VisualizationBase.__init__(self, identifier)  # process the identifier
        self.geometry = geometry
        self._min_point = None
        self._max_point = None

    @classmethod
    def from_dict(cls, data):
        """Create an ContextGeometry from a dictionary.

        Args:
            data: A python dictionary in the following format

        .. code-block:: python

            {
            "type": "ContextGeometry",
            "identifier": "",  # unique object identifier
            "geometry": []  # list of ladybug-display geometry objects
            }
        """
        # check the type key
        assert data['type'] == 'ContextGeometry', \
            'Expected ContextGeometry, Got {}.'.format(data['type'])
        # re-serialize the object
        geos = tuple(dict_to_object(geo) for geo in data['geometry'])
        new_obj = cls(data['identifier'], geos)
        if 'display_name' in data and data['display_name'] is not None:
            new_obj.display_name = data['display_name']
        if 'user_data' in data and data['user_data'] is not None:
            new_obj.user_data = data['user_data']
        return new_obj

    @property
    def geometry(self):
        """Get or set a tuple of ladybug_display geometry objects for context.

        When setting this property, it is also acceptable to include raw
        ladybug_geometry objects in the list and they will automatically be
        converted into a wireframe representation as a ladybug-display object.
        """
        return self._geometry

    @geometry.setter
    def geometry(self, value):
        assert isinstance(value, (list, tuple)), 'Expected list or tuple for' \
            ' ContextGeometry geometry. Got {}.'.format(type(value))
        processed_value = []
        for geo in value:
            if isinstance(geo, DISPLAY_UNION):
                processed_value.append(geo)
            elif isinstance(geo, GEOMETRY_UNION):
                processed_value.append(self.geometry_to_wireframe(geo))
            else:
                raise ValueError(
                    'Expected ladybug-geometry or ladybug-display object for '
                    'ContextGeometry. Got {}.'.format(type(geo)))
        self._geometry = tuple(processed_value)

    @property
    def min_point(self):
        """A Point3D for the minimum bounding box vertex around all of the geometry."""
        if self._min_point is None:
            self._calculate_min_max()
        return self._min_point

    @property
    def max_point(self):
        """A Point3D for the maximum bounding box vertex around all of the geometry."""
        if self._max_point is None:
            self._calculate_min_max()
        return self._max_point

    def to_dict(self):
        """Get ContextGeometry as a dictionary."""
        base = {
            'type': 'ContextGeometry',
            'identifier': self.identifier,
            'geometry': [geo.to_dict() for geo in self.geometry]
        }
        if self._display_name is not None:
            base['display_name'] = self.display_name
        if self.user_data is not None:
            base['user_data'] = self.user_data
        return base

    @staticmethod
    def geometry_to_wireframe(geometry):
        """Convert a raw ladybug-geometry object into a wireframe ladybug-display object.

        Args:
            geometry: A raw ladybug-geometry object to be converted to a wireframe
                ladybug-display object.
        """
        conv_info = ContextGeometry.WIREFRAME_MAP[geometry.__class__]
        new_class, wire_args = conv_info[0], list(conv_info[1:])
        wire_args.insert(0, geometry)
        return new_class(*wire_args)

    def _calculate_min_max(self):
        """Calculate maximum and minimum Point3D for this object."""
        lb_geos = [d_geo.geometry for d_geo in self.geometry]
        self._min_point, self._max_point = bounding_box(lb_geos)

    def __len__(self):
        """Return number of geometries on the object."""
        return len(self.geometry)

    def __getitem__(self, key):
        """Return one of the geometries."""
        return self.geometry[key]

    def __iter__(self):
        """Iterate through the geometries."""
        return iter(self.geometry)

    def ToString(self):
        """Overwrite .NET ToString."""
        return self.__repr__()

    def __repr__(self):
        """AnalysisGeometry representation."""
        return 'Context Geometry: {}'.format(self.display_name)


class AnalysisGeometry(_VisualizationBase):
    """An object where multiple data streams correspond to the same geometry.

    Args:
        identifier: Text string for a unique object ID. Must be less than 100
            characters and not contain spaces or special characters.
        geometry: A list of ladybug-geometry objects that is aligned with the values in
            the input data_sets. The length of this list should usually be equal to the
            total number of values in each data_set, indicating that each geometry
            gets a single color. Alternatively, if all of the geometry objects are
            meshes, the number of values in the data can be equal to the total number
            of faces across the meshes or the total number of vertices across the meshes.
        data_sets: A list of VisualizationData objects representing the data sets
            that are associated with the input geometry.
        active_data: An integer to denote which of the input data_sets should be
            displayed by default. (Default: 0).
        display_mode: Text to indicate the display mode (surface, wireframe, etc.).
            Choose from the following. (Default: Surface).

            * Surface
            * SurfaceWithEdges
            * Wireframe
            * Points

    Properties:
        * identifier
        * display_name
        * geometry
        * data_sets
        * active_data
        * display_mode
        * min_point
        * max_point
        * matching_method
        * user_data
    """
    __slots__ = (
        '_geometry', '_data_sets', '_active_data', '_display_mode',
        '_min_point', '_max_point', '_possible_lengths', '_matching_method')

    def __init__(self, identifier, geometry, data_sets,
                 active_data=0, display_mode='Surface'):
        """Initialize AnalysisGeometry."""
        _VisualizationBase.__init__(self, identifier)  # process the identifier
        if not isinstance(geometry, tuple):
            geometry = tuple(geometry)
        if not isinstance(data_sets, tuple):
            data_sets = tuple(data_sets)
        self._possible_lengths = self._possible_data_lengths(geometry)
        self._matching_method = None
        for dat in data_sets:
            self._check_data_set(dat)
        self._geometry = geometry
        self._data_sets = data_sets
        self.active_data = active_data
        self.display_mode = display_mode
        self._min_point = None
        self._max_point = None

    @classmethod
    def from_dict(cls, data):
        """Create an AnalysisGeometry from a dictionary.

        Args:
            data: A python dictionary in the following format

        .. code-block:: python

            {
            "type": "AnalysisGeometry",
            "identifier": "",  # unique object identifier
            "geometry": [],  # list of geometry objects
            "data_sets": [],  # list of data sets associated with the geometry
            "active_data": 0,  # integer for the index of the active data set
            "display_mode": "Surface"  # text for the display mode of the data
            }
        """
        # check the type key
        assert data['type'] == 'AnalysisGeometry', \
            'Expected AnalysisGeometry, Got {}.'.format(data['type'])
        # re-serialize the geometry and data sets
        geos = tuple(geometry_dict_to_object(geo) for geo in data['geometry'])
        dts = tuple(VisualizationData.from_dict(dt) for dt in data['data_sets'])
        # re-serialize the data type and unit
        act_dt = data['active_data'] if 'active_data' in data else 0
        d_mode = data['display_mode'] if 'display_mode' in data else 'Surface'
        new_obj = cls(data['identifier'], geos, dts, act_dt, d_mode)
        if 'display_name' in data and data['display_name'] is not None:
            new_obj.display_name = data['display_name']
        if 'user_data' in data and data['user_data'] is not None:
            new_obj.user_data = data['user_data']
        return new_obj

    @property
    def geometry(self):
        """Get a tuple of geometry objects assigned to this AnalysisGeometry."""
        return self._geometry

    @property
    def data_sets(self):
        """Get a tuple of VisualizationData assigned to the AnalysisGeometry."""
        return self._data_sets

    @property
    def active_data(self):
        """Get or set an integer for the index of data set that is actively displaying.
        """
        return self._active_data

    @active_data.setter
    def active_data(self, value):
        self._active_data = int_in_range(value, 0, len(self._data_sets), 'active_data')

    @property
    def display_mode(self):
        """Get or set text to indicate the display mode."""
        return self._display_mode

    @display_mode.setter
    def display_mode(self, value):
        clean_input = value.lower()
        for key in DISPLAY_MODES:
            if key.lower() == clean_input:
                value = key
                break
        else:
            raise ValueError(
                'display_mode {} is not recognized.\nChoose from the '
                'following:\n{}'.format(value, DISPLAY_MODES))
        self._display_mode = value

    @property
    def min_point(self):
        """A Point3D for the minimum bounding box vertex around all of the geometry."""
        if self._min_point is None:
            self._calculate_min_max()
        return self._min_point

    @property
    def max_point(self):
        """A Point3D for the maximum bounding box vertex around all of the geometry."""
        if self._max_point is None:
            self._calculate_min_max()
        return self._max_point

    @property
    def matching_method(self):
        """Get text for the method by which the data is matched to the geometry.

        This will be one of the following.

        * geometries - One value is assigned for each geometry
        * faces - One value is assigned per each face of the Mesh
        * vertices - One value is assigned per each vertex of the Mesh
        """
        return self._matching_method

    @property
    def user_data(self):
        """Get or set an optional dictionary for additional meta data for this object.

        This will be None until it has been set. All keys and values of this
        dictionary should be of a standard Python type to ensure correct
        serialization of the object to/from JSON (eg. str, float, int, list, dict)
        """
        return self._user_data

    @user_data.setter
    def user_data(self, value):
        if value is not None:
            assert isinstance(value, dict), 'Expected dictionary for ' \
                'object user_data. Got {}.'.format(type(value))
        self._user_data = value

    def add_data_set(self, data, insert_index=None):
        """Add a data set to this AnalysisGeometry object.

        Args:
            data: A VisualizationData object to be added to this AnalysisGeometry.
            insert_index: An integer for the index at which the data should be
                inserted. If None, the data will be appended to the end. (Default: None).
        """
        self._check_data_set(data)
        if insert_index is None:
            self._data_sets = self._data_sets + (data,)
        else:
            ds_list = list(self._data_sets)
            ds_list.insert(insert_index, data)
            self._data_sets = tuple(ds_list)

    def remove_data_set(self, data_index):
        """Remove a data set from this AnalysisGeometry object.

        Args:
            data_index: An integer for the data index to be removed.
        """
        assert data_index < len(self._data_sets), 'data_index ({}) must be less than ' \
            'the number of items in the data_sets ({}).'.format(
                data_index, len(self._data_sets))
        ds_list = list(self._data_sets)
        ds_list.pop(data_index)
        self._data_sets = tuple(ds_list)
        if self._active_data == len(self._data_sets):
            self._active_data = len(self._data_sets) - 1

    def graphic_container(self, data_index=None, min_point=None, max_point=None):
        """Get a Ladybug GraphicContainer object, which can be used to draw legends.

        Args:
            data_index: Integer for the index of the data set for which a
                GraphicContainer will be returned. If None, the active_data set
                will be used. (Default: None).
            min_point: An optional Point3D to denote the minimum bounding box
                for the graphic container. If None, this object's own min_point
                will be used, which corresponds to the bounding box around
                the geometry. (Default: None).
            max_point: An optional Point3D to denote the maximum bounding box
                for the graphic container. If None, this object's own max_point
                will be used, which corresponds to the bounding box around
                the geometry. (Default: None).
        """
        # ensure that min and max points always make sense
        min_point = self.min_point if min_point is None else min_point
        max_point = self.max_point if max_point is None else max_point
        # return the Graphic Container for the correct data set
        data_index = self.active_data if data_index is None else data_index
        return self.data_sets[data_index].graphic_container(min_point, max_point)

    def to_dict(self):
        """Get AnalysisGeometry as a dictionary."""
        base = {
            'type': 'AnalysisGeometry',
            'identifier': self.identifier,
            'geometry': [geo.to_dict() for geo in self.geometry],
            'data_sets': [ds.to_dict() for ds in self.data_sets],
            'active_data': self.active_data,
            'display_mode': self.display_mode
        }
        if self._display_name is not None:
            base['display_name'] = self.display_name
        if self.user_data is not None:
            base['user_data'] = self.user_data
        return base

    def _check_data_set(self, data_set):
        """Check that a data set is compatible with the geometry."""
        assert isinstance(data_set, VisualizationData), 'Expected VisualizationData ' \
            'for AnalysisGeometry. Got {}.'.format(type(data_set))
        if self._matching_method is None:  # first data set to be matched
            pl = self._possible_lengths
            if pl[1] == 0 and pl[2] == 0:
                assert len(data_set) == pl[0], 'Expected number of data set values ' \
                    '({}) to align with the number of geometries ({}).'.format(
                        len(data_set.values), pl[0])
            else:
                assert len(data_set) in pl, 'Expected number of data set values ' \
                    '({}) to align with the number of geometries ({}), the number of ' \
                    'geometry faces ({}), or the number of geometry vertices ' \
                    '({}).'.format(len(data_set.values), pl[0], pl[1], pl[2])
            self._matching_method = self._matching_type(data_set)[0]
        else:
            assert self._matching_type(data_set)[0] == self._matching_method, \
                'Expected number of data set values ({}) to align with the number ' \
                'of {} ({}).'.format(len(data_set.values), self._matching_method,
                                     self._matching_type(data_set)[1])

    def _matching_type(self, dat_set):
        """Get text and number of values for the method by which data and geometry match.

        Args:
            dat_set: A data set which will have its length evaluated in relation to
                this object's geometry.
        """
        geo_len, face_len, vert_len = self._possible_lengths
        if len(dat_set) == geo_len:
            return 'geometries', geo_len
        if len(dat_set) == face_len:
            return 'faces', face_len
        if len(dat_set) == vert_len:
            return 'vertices', vert_len

    def _calculate_min_max(self):
        """Calculate maximum and minimum Point3D for this object."""
        self._min_point, self._max_point = bounding_box(self.geometry)

    @staticmethod
    def _possible_data_lengths(geometry):
        """Get the acceptable data lengths given the input geometry."""
        geo_count_0, geo_count_1, geo_count_2 = len(geometry), 0, 0
        for geo in geometry:
            if isinstance(geo, (Mesh2D, Mesh3D)):
                geo_count_1 += len(geo.faces)
                geo_count_2 += len(geo.vertices)
            else:
                assert isinstance(geo, GEOMETRY_UNION), 'Expected ladybug geometry ' \
                    'object for AnalysisGeometry. Got {}.'.format(type(geo))
        return (geo_count_0, geo_count_1, geo_count_2)

    def __len__(self):
        """Return number of data sets on the object."""
        return len(self.data_sets)

    def __getitem__(self, key):
        """Return one of the data sets."""
        return self.data_sets[key]

    def __iter__(self):
        """Iterate through the data sets."""
        return iter(self.data_sets._values)

    def ToString(self):
        """Overwrite .NET ToString."""
        return self.__repr__()

    def __repr__(self):
        """AnalysisGeometry representation."""
        return 'Analysis Geometry: {}'.format(self.display_name)


class VisualizationMetaData(object):
    """Represents the metadata for visualization with legend parameters and data type.

    Args:
        legend_parameters: An Optional LegendParameters object to override default
            parameters of the legend. None indicates that default legend parameters
            will be used. (Default: None).
        data_type: Optional DataType from the ladybug datatype subpackage (ie.
            Temperature()) , which will be used to assign default legend properties.
            If None, the legend associated with this object will contain no units
            unless a unit below is specified. (Default: None).
        unit: Optional text string for the units of the values. (ie. "C"). If None
            or empty, the default units of the data_type will be used. If no data
            type is specified in this case, this will simply be an empty
            string. (Default: None).

    Properties:
        * legend_parameters
        * data_type
        * user_data
    """
    __slots__ = ('_legend_parameters', '_data_type', '_unit', '_user_data')

    def __init__(self, legend_parameters=None, data_type=None, unit=None):
        """Initialize VisualizationMetaData."""
        self._legend_parameters = legend_parameters
        self._user_data = None

        # set default legend parameters based on input data_type and unit
        self._data_type = data_type
        self._unit = unit
        if data_type is not None:
            assert isinstance(data_type, DataTypeBase), \
                'data_type should be a ladybug DataType. Got {}'.format(type(data_type))
            if self.legend_parameters.is_title_default:
                unit = unit if unit else data_type.units[0]
                data_type.is_unit_acceptable(unit)
                self.legend_parameters.title = unit if \
                    self.legend_parameters.vertical \
                    else '{} ({})'.format(data_type.name, unit)
            if data_type.unit_descr is not None and \
                    self.legend_parameters.ordinal_dictionary is None and not \
                    isinstance(self.legend_parameters, LegendParametersCategorized):
                self.legend_parameters.ordinal_dictionary = data_type.unit_descr
                sorted_keys = sorted(data_type.unit_descr.keys())
                if self.legend.is_min_default:
                    self.legend_parameters._min = sorted_keys[0]
                if self.legend.is_max_default:
                    self.legend_parameters._max = sorted_keys[-1]
                assert self.legend_parameters._min <= self.legend_parameters._max, \
                    'Legend min is greater than legend max. {} > {}.'.format(
                        self.legend_parameters._min, self.legend_parameters._max)
                if self.legend_parameters.is_segment_count_default:
                    try:  # try to set the number of segments to align with ordinal text
                        min_i = sorted_keys.index(self.legend_parameters.min)
                        max_i = sorted_keys.index(self.legend_parameters.max)
                        self.legend_parameters.segment_count = \
                            len(sorted_keys[min_i:max_i + 1])
                    except IndexError:
                        pass
        elif unit and self.legend_parameters.is_title_default:
            assert isinstance(unit, str), \
                'Expected string for unit. Got {}.'.format(type(unit))
            self.legend_parameters.title = unit

    @classmethod
    def from_dict(cls, data):
        """Create VisualizationMetaData from a dictionary.

        Args:
            data: A python dictionary in the following format

        .. code-block:: python

            {
            "type": "VisualizationMetaData",
            "legend_parameters": {},  # optional LegendParameter specification
            "data_type": {},  # optional DataType object
            "unit": "C"  # optional text for the units
            }
        """
        # check the type key
        assert data['type'] == 'VisualizationMetaData', \
            'Expected VisualizationMetaData, Got {}.'.format(data['type'])
        # re-serialize the legend parameters
        legend_parameters = None
        if 'legend_parameters' in data and data['legend_parameters'] is not None:
            if data['legend_parameters']['type'] == 'LegendParametersCategorized':
                legend_parameters = LegendParametersCategorized.from_dict(
                    data['legend_parameters'])
            else:
                legend_parameters = LegendParameters.from_dict(data['legend_parameters'])
        # re-serialize the data type and unit
        data_type = None
        if 'data_type' in data and data['data_type'] is not None:
            data_type = DataTypeBase.from_dict(data['data_type'])
        unit = data['unit'] if 'unit' in data else None
        new_obj = cls(legend_parameters, data_type, unit)
        if 'user_data' in data and data['user_data'] is not None:
            new_obj.user_data = data['user_data']
        return new_obj

    @property
    def legend_parameters(self):
        """Get the legend parameters assigned to this data set."""
        return self._legend_parameters

    @property
    def data_type(self):
        """Get the data_type input to this object (if it exists)."""
        return self._data_type

    @property
    def unit(self):
        """Get the unit input to this object (if it exists)."""
        return self._unit

    @property
    def user_data(self):
        """Get or set an optional dictionary for additional meta data for this object.

        This will be None until it has been set. All keys and values of this
        dictionary should be of a standard Python type to ensure correct
        serialization of the object to/from JSON (eg. str, float, int, list, dict)
        """
        return self._user_data

    @user_data.setter
    def user_data(self, value):
        if value is not None:
            assert isinstance(value, dict), 'Expected dictionary for ' \
                'object user_data. Got {}.'.format(type(value))
        self._user_data = value

    def to_dict(self):
        """Get visualization data as a dictionary."""
        base = {
            'type': 'VisualizationMetaData'
        }
        if self._legend_parameters is not None:
            base['legend_parameters'] = self._legend_parameters.to_dict()
        if self.data_type is not None:
            base['data_type'] = self.data_type.to_dict()
        if self.unit:
            base['unit'] = self.unit
        if self.user_data is not None:
            base['user_data'] = self.user_data
        return base

    def ToString(self):
        """Overwrite .NET ToString."""
        return self.__repr__()

    def __repr__(self):
        """VisualizationMetaData representation."""
        return 'Visualization MetaData'


class VisualizationData(VisualizationMetaData):
    """Represents a data set for visualization with legend parameters and data type.

    Args:
        values: A list of numerical values that will be used to generate the
            visualization colors.
        legend_parameters: An Optional LegendParameters object to override default
            parameters of the legend. None indicates that default legend parameters
            will be used. (Default: None).
        data_type: Optional DataType from the ladybug datatype subpackage (ie.
            Temperature()) , which will be used to assign default legend properties.
            If None, the legend associated with this object will contain no units
            unless a unit below is specified. (Default: None).
        unit: Optional text string for the units of the values. (ie. "C"). If None
            or empty, the default units of the data_type will be used. If no data
            type is specified in this case, this will simply be an empty
            string. (Default: None).

    Properties:
        * values
        * legend_parameters
        * data_type
        * user_data
    """
    __slots__ = ('_legend', '_legend_parameters', '_data_type', '_unit', '_user_data')

    def __init__(self, values, legend_parameters=None, data_type=None, unit=None):
        """Initialize VisualizationData."""
        # set up the legend using the values and legend parameters
        VisualizationMetaData.__init__(self, legend_parameters, data_type, unit)
        self._legend = Legend(values, legend_parameters)

    @classmethod
    def from_dict(cls, data):
        """Create VisualizationData from a dictionary.

        Args:
            data: A python dictionary in the following format

        .. code-block:: python

            {
            "type": "VisualizationData",
            "values": [0, 10],
            "legend_parameters": {},  # optional LegendParameter specification
            "data_type": {},  # optional DataType object
            "unit": "C"  # optional text for the units
            }
        """
        # check the type key
        assert data['type'] == 'VisualizationData', \
            'Expected VisualizationData, Got {}.'.format(data['type'])
        # re-serialize the legend parameters
        legend_parameters = None
        if 'legend_parameters' in data and data['legend_parameters'] is not None:
            if data['legend_parameters']['type'] == 'LegendParametersCategorized':
                legend_parameters = LegendParametersCategorized.from_dict(
                    data['legend_parameters'])
            else:
                legend_parameters = LegendParameters.from_dict(data['legend_parameters'])
        # re-serialize the data type and unit
        data_type = None
        if 'data_type' in data and data['data_type'] is not None:
            data_type = DataTypeBase.from_dict(data['data_type'])
        unit = data['unit'] if 'unit' in data else None
        new_obj = cls(data['values'], legend_parameters, data_type, unit)
        if 'user_data' in data and data['user_data'] is not None:
            new_obj.user_data = data['user_data']
        return new_obj

    @property
    def values(self):
        """Get the values assigned to the data set."""
        return self._legend.values

    @property
    def legend(self):
        """Get the legend assigned to this data set."""
        return self._legend

    @property
    def value_colors(self):
        """Get a List of colors associated with the assigned values."""
        return self._legend.value_colors

    def graphic_container(self, min_point, max_point):
        """Get a Ladybug GraphicContainer object, which can be used to a draw legend.

        Args:
            min_point: An optional Point3D to denote the minimum bounding box
                for the graphic container.
            max_point: An optional Point3D to denote the maximum bounding box
                for the graphic container.
        """
        return GraphicContainer(
            self.values, min_point, max_point, self._legend_parameters,
            self._data_type, self._unit)

    def to_dict(self):
        """Get visualization data as a dictionary."""
        base = {
            'type': 'VisualizationData',
            'values': self.values
        }
        if self._legend_parameters is not None:
            base['legend_parameters'] = self._legend_parameters.to_dict()
        if self.data_type is not None:
            base['data_type'] = self.data_type.to_dict()
        if self.unit:
            base['unit'] = self.unit
        if self.user_data is not None:
            base['user_data'] = self.user_data
        return base

    def __len__(self):
        """Return length of values on the object."""
        return len(self._legend._values)

    def __getitem__(self, key):
        """Return one of the values."""
        return self._legend._values[key]

    def __iter__(self):
        """Iterate through the values."""
        return iter(self._legend._values)

    def ToString(self):
        """Overwrite .NET ToString."""
        return self.__repr__()

    def __repr__(self):
        """VisualizationData representation."""
        return 'Visualization Data ({} values)'.format(len(self))



# coding=utf-8
from ladybug_geometry.geometry2d.pointvector import Point2D
from ladybug_geometry.geometry2d.mesh import Mesh2D
from ladybug.color import Color
from ladybug_display.geometry2d.mesh import DisplayMesh2D


def test_display_mesh2d_init():
    """Test the initialization of DisplayMesh2D objects with two faces."""
    grey = Color(100, 100, 100)
    pts = (Point2D(0, 0), Point2D(0, 2), Point2D(2, 2),
           Point2D(2, 0), Point2D(4, 0))
    mesh = DisplayMesh2D(Mesh2D(pts, [(0, 1, 2, 3), (2, 3, 4)]), [grey])

    assert mesh.colors == (grey,)
    assert mesh.display_mode == 'Surface'
    assert len(mesh.vertices) == 5
    assert len(mesh.faces) == 2
    assert mesh.area == 6
    assert mesh.min == Point2D(0, 0)
    assert mesh.max == Point2D(4, 2)
    assert mesh.center == Point2D(2, 1)
    assert len(mesh.face_areas) == 2
    assert mesh.face_areas[0] == 4
    assert mesh.face_areas[1] == 2
    assert len(mesh.face_centroids) == 2
    assert mesh.face_centroids[0] == Point2D(1, 1)

    blue = Color(0, 0, 100)
    mesh.colors = (blue,)
    mesh.display_mode = 'Wireframe'
    assert mesh.colors == (blue,)
    assert mesh.display_mode == 'Wireframe'


def test_mesh2d_to_from_dict():
    """Test the to/from dict of Mesh2D objects."""
    grey = Color(100, 100, 100)
    pts = (Point2D(0, 0), Point2D(0, 2), Point2D(2, 2), Point2D(2, 0))
    mesh = DisplayMesh2D(Mesh2D(pts, [(0, 1, 2, 3)]), [grey])
    mesh.display_mode = 'Wireframe'
    mesh_dict = mesh.to_dict()
    new_mesh = DisplayMesh2D.from_dict(mesh_dict)
    assert isinstance(new_mesh, DisplayMesh2D)
    print(new_mesh.colors)
    assert new_mesh.to_dict() == mesh_dict



# coding=utf-8
from ladybug_geometry.geometry3d.pointvector import Point3D, Vector3D
from ladybug_geometry.geometry3d.plane import Plane
from ladybug_geometry.geometry3d.face import Face3D
from ladybug.color import Color
from ladybug_display.geometry3d.face import DisplayFace3D


def test_display_face3d_init():
    """Test the initialization of DisplayFace3D objects and basic properties."""
    grey = Color(100, 100, 100)
    pts = (Point3D(0, 0, 2), Point3D(0, 2, 2), Point3D(2, 2, 2), Point3D(2, 0, 2))
    plane = Plane(Vector3D(0, 0, 1), Point3D(0, 0, 2))
    face = DisplayFace3D(Face3D(pts, plane), grey)
    str(face)  # test the string representation of the face

    assert face.color == grey
    assert face.display_mode == 'Surface'
    assert isinstance(face.vertices, tuple)
    assert len(face.vertices) == 4
    for point in face.vertices:
        assert isinstance(point, Point3D)
    assert face.area == 4
    assert face.perimeter == 8
    assert round(face.altitude, 3) == round(90, 3)
    assert round(face.azimuth, 3) == 0

    blue = Color(0, 0, 100)
    face.color = blue
    face.display_mode = 'Wireframe'
    assert face.color == blue
    assert face.display_mode == 'Wireframe'


def test_face3d_to_from_dict():
    """Test the to/from dict of Face3D objects."""
    grey = Color(100, 100, 100)
    pts = (Point3D(0, 0, 2), Point3D(0, 2, 2), Point3D(2, 2, 2), Point3D(2, 0, 2))
    plane = Plane(Vector3D(0, 0, 1), Point3D(0, 0, 2))
    face = DisplayFace3D(Face3D(pts, plane), grey)
    face.display_mode = 'Wireframe'
    face_dict = face.to_dict()
    new_face = DisplayFace3D.from_dict(face_dict)
    assert isinstance(new_face, DisplayFace3D)
    assert new_face.to_dict() == face_dict

    bound_pts = [Point3D(0, 0), Point3D(4, 0), Point3D(4, 4), Point3D(0, 4)]
    hole_pts = [Point3D(1, 1), Point3D(3, 1), Point3D(3, 3), Point3D(1, 3)]
    face = DisplayFace3D(Face3D(bound_pts, None, [hole_pts]), grey)
    face_dict = face.to_dict()
    new_face = DisplayFace3D.from_dict(face_dict)
    assert isinstance(new_face, DisplayFace3D)
    assert new_face.to_dict() == face_dict



"""A cylinder that can be displayed in 3D space."""
from ladybug_geometry.geometry3d.cylinder import Cylinder
from ladybug.color import Color

from ._base import _SingleColorModeBase3D


class DisplayCylinder(_SingleColorModeBase3D):
    """A cylinder in 3D space with display properties.

    Args:
        geometry: A ladybug-geometry Cylinder.
        color: A ladybug Color object. If None, a default black color will be
            used. (Default: None).
        display_mode: Text to indicate the display mode (surface, wireframe, etc.).
            Choose from the following. (Default: Surface).

            * Surface
            * SurfaceWithEdges
            * Wireframe
            * Points

    Properties:
        * geometry
        * color
        * display_mode
        * center
        * axis
        * height
        * radius
        * area
        * volume
        * user_data
    """
    __slots__ = ()

    def __init__(self, geometry, color=None, display_mode='Surface'):
        """Initialize base with shade object."""
        assert isinstance(geometry, Cylinder), '\
            Expected ladybug_geometry Cylinder. Got {}'.format(type(geometry))
        _SingleColorModeBase3D.__init__(self, geometry, color, display_mode)

    @classmethod
    def from_dict(cls, data):
        """Initialize a DisplayCylinder from a dictionary.

        Args:
            data: A dictionary representation of an DisplayCylinder object.
        """
        assert data['type'] == 'DisplayCylinder', \
            'Expected DisplayCylinder dictionary. Got {}.'.format(data['type'])
        color = Color.from_dict(data['color']) if 'color' in data and data['color'] \
            is not None else None
        d_mode = data['display_mode'] if 'display_mode' in data and \
            data['display_mode'] is not None else 'Surface'
        geo = cls(Cylinder.from_dict(data['geometry']), color, d_mode)
        if 'user_data' in data and data['user_data'] is not None:
            geo.user_data = data['user_data']
        return geo

    @property
    def center(self):
        """Get a Point3D for the center of the bottom of the cylinder."""
        return self._geometry.center

    @property
    def axis(self):
        """Get a Vector3D for the axis of the cylinder."""
        return self._geometry.axis

    @property
    def height(self):
        """Get a number for the height of the cylinder."""
        return self._geometry.height

    @property
    def radius(self):
        """Get the radius of the cylinder."""
        return self._geometry.radius

    @property
    def area(self):
        """Get the surface area of the cylinder."""
        return self._geometry.area

    @property
    def volume(self):
        """Get the volume of the cylinder."""
        return self._geometry.volume

    def to_dict(self):
        """Return DisplayCylinder as a dictionary."""
        base = {'type': 'DisplayCylinder'}
        base['geometry'] = self._geometry.to_dict()
        base['color'] = self.color.to_dict()
        base['display_mode'] = self.display_mode
        if self.user_data is not None:
            base['user_data'] = self.user_data
        return base

    def __copy__(self):
        new_g = DisplayCylinder(self.geometry, self.color, self.display_mode)
        new_g._user_data = None if self.user_data is None else self.user_data.copy()
        return new_g

    def __repr__(self):
        return 'DisplayCylinder: {}'.format(self.geometry)



# coding=utf-8
from ladybug_geometry.geometry3d.pointvector import Point3D, Vector3D
from ladybug_geometry.geometry3d.ray import Ray3D
from ladybug.color import Color
from ladybug_display.geometry3d.ray import DisplayRay3D


def test_display_ray3d_init():
    """Test the initialization of Ray3D objects and basic properties."""
    grey = Color(100, 100, 100)
    pt = Point3D(2, 0, 2)
    vec = Vector3D(0, 2, 0)
    ray = DisplayRay3D(Ray3D(pt, vec), grey)
    str(ray)  # test the string representation of the ray

    assert ray.color == grey
    assert ray.p == Point3D(2, 0, 2)
    assert ray.v == Vector3D(0, 2, 0)

    ray.reverse()
    assert ray.p == Point3D(2, 0, 2)
    assert ray.v == Vector3D(0, -2, 0)


def test_ray3_to_from_dict():
    """Test the to/from dict of Ray3D objects."""
    grey = Color(100, 100, 100)
    pt = Point3D(2, 0, 2)
    vec = Vector3D(0, 2, 0)
    ray = DisplayRay3D(Ray3D(pt, vec), grey)
    ray_dict = ray.to_dict()
    new_ray = DisplayRay3D.from_dict(ray_dict)
    assert isinstance(new_ray, DisplayRay3D)
    assert new_ray.to_dict() == ray_dict



# coding=utf-8
import pytest

from ladybug_geometry.geometry2d import Vector2D
from ladybug.color import Color
from ladybug_display.geometry2d.vector import DisplayVector2D


def test_display_vector2d_init():
    """Test the initialization of DisplayVector2D objects and basic properties."""
    grey = Color(100, 100, 100)
    vec = DisplayVector2D(Vector2D(0, 2), grey)
    str(vec)  # test the string representation of the vector

    assert vec.x == 0
    assert vec.y == 2
    assert vec.color == grey
    assert vec[0] == 0
    assert vec[1] == 2
    assert vec.magnitude == 2

    pt_tuple = tuple(i for i in vec)
    assert pt_tuple == (0, 2)

    vec.normalize()
    assert vec.x == 0
    assert vec.magnitude == 1

    norm_vec = vec * 3
    assert norm_vec.magnitude == 3


def test_display_vector2d_to_from_dict():
    """Test the initialization of DisplayVector2D objects and basic properties."""
    grey = Color(100, 100, 100)
    vec = DisplayVector2D(Vector2D(0, 2), grey)
    vec_dict = vec.to_dict()
    new_vec = DisplayVector2D.from_dict(vec_dict)
    assert isinstance(new_vec, DisplayVector2D)
    assert new_vec.to_dict() == vec_dict


def test_display_vector2d_angle():
    """Test the methods that get the angle between DisplayVector2D objects."""
    grey = Color(100, 100, 100)
    vec_1 = DisplayVector2D(Vector2D(0, 2), grey)
    vec_2 = DisplayVector2D(Vector2D(2, 0), grey)
    vec_3 = DisplayVector2D(Vector2D(0, -2), grey)
    vec_4 = DisplayVector2D(Vector2D(-2, 0), grey)
    assert vec_1.angle(vec_2) == pytest.approx(90, rel=1e-3)
    assert vec_1.angle(vec_3) == pytest.approx(180, rel=1e-3)
    assert vec_1.angle(vec_4) == pytest.approx(90, rel=1e-3)
    assert vec_1.angle(vec_1) == pytest.approx(0, rel=1e-3)



"""A single planar face that can be displayed in 3D space."""
import math

from ladybug_geometry.geometry3d.face import Face3D
from ladybug.color import Color

from ._base import _SingleColorModeBase3D


class DisplayFace3D(_SingleColorModeBase3D):
    """A single planar face in 3D space with display properties.

    Args:
        geometry: A ladybug-geometry Face3D.
        color: A ladybug Color object. If None, a default black color will be
            used. (Default: None).
        display_mode: Text to indicate the display mode (surface, wireframe, etc.).
            Choose from the following. (Default: Surface).

            * Surface
            * SurfaceWithEdges
            * Wireframe
            * Points

    Properties:
        * geometry
        * color
        * display_mode
        * vertices
        * normal
        * center
        * area
        * perimeter
        * min
        * max
        * altitude
        * azimuth
        * user_data
    """
    __slots__ = ()

    def __init__(self, geometry, color=None, display_mode='Surface'):
        """Initialize object."""
        assert isinstance(geometry, Face3D), '\
            Expected ladybug_geometry Face3D. Got {}'.format(type(geometry))
        _SingleColorModeBase3D.__init__(self, geometry, color, display_mode)

    @classmethod
    def from_dict(cls, data):
        """Initialize a DisplayFace3D from a dictionary.

        Args:
            data: A dictionary representation of an DisplayFace3D object.
        """
        assert data['type'] == 'DisplayFace3D', \
            'Expected DisplayFace3D dictionary. Got {}.'.format(data['type'])
        color = Color.from_dict(data['color']) if 'color' in data and data['color'] \
            is not None else None
        d_mode = data['display_mode'] if 'display_mode' in data and \
            data['display_mode'] is not None else 'Surface'
        geo = cls(Face3D.from_dict(data['geometry']), color, d_mode)
        if 'user_data' in data and data['user_data'] is not None:
            geo.user_data = data['user_data']
        return geo

    @property
    def vertices(self):
        """Get a list of vertices for the face (in counter-clockwise order)."""
        return self._geometry.vertices

    @property
    def normal(self):
        """Get a Vector3D for the direction in which the face is pointing.
        """
        return self._geometry.normal

    @property
    def center(self):
        """Get a Point3D for the center of the face.

        Note that this is the center of the bounding rectangle around this geometry
        and not the area centroid.
        """
        return self._geometry.center

    @property
    def area(self):
        """Get the area of the face."""
        return self._geometry.area

    @property
    def perimeter(self):
        """Get the perimeter of the face. This includes the length of holes in the face.
        """
        return self._geometry.perimeter

    @property
    def min(self):
        """Get a Point3D for the minimum of the bounding box around the object."""
        return self._geometry.min

    @property
    def max(self):
        """Get a Point3D for the maximum of the bounding box around the object."""
        return self._geometry.max

    @property
    def altitude(self):
        """Get the altitude of the geometry between +90 (up) and -90 (down)."""
        return math.degrees(self._geometry.altitude)

    @property
    def azimuth(self):
        """Get the azimuth of the geometry, between 0 and 360.

        Given Y-axis as North, 0 = North, 90 = East, 180 = South, 270 = West
        This will be zero if the Face3D is perfectly horizontal.
        """
        return math.degrees(self._geometry.azimuth)

    def to_dict(self, include_plane=True):
        """Return DisplayFace3D as a dictionary.

        Args:
            include_plane: Boolean to note wether the plane of the Face3D should be
                included in the output. This can preserve the orientation of the
                X/Y axes of the plane but is not required and can be removed to
                keep the dictionary smaller. (Default: True).
        """
        base = {'type': 'DisplayFace3D'}
        base['geometry'] = self._geometry.to_dict(include_plane)
        base['color'] = self.color.to_dict()
        base['display_mode'] = self.display_mode
        if self.user_data is not None:
            base['user_data'] = self.user_data
        return base

    def __copy__(self):
        new_g = DisplayFace3D(self.geometry, self.color, self.display_mode)
        new_g._user_data = None if self.user_data is None else self.user_data.copy()
        return new_g

    def __repr__(self):
        return 'DisplayFace3D: {}'.format(self.geometry)



# coding: utf-8
"""Base classes for all 3D geometry objects."""
import math

from ladybug.color import Color

from ladybug_display._base import _DisplayBase, DISPLAY_MODES, LINE_TYPES
from ladybug_display.altnumber import default
from ladybug_display.typing import float_positive


class _DisplayBase3D(_DisplayBase):
    """A base class for all 3D ladybug-display geometry objects.

    Args:
        geometry: A ladybug-geometry object.

    Properties:
        * geometry
        * user_data
    """
    __slots__ = ()

    def __init__(self, geometry):
        """Initialize base with shade object."""
        _DisplayBase.__init__(self, geometry)

    def move(self, moving_vec):
        """Move this geometry along a vector.

        Args:
            moving_vec: A ladybug_geometry Vector with the direction and distance
                to move the geometry.
        """
        self._geometry = self.geometry.move(moving_vec)

    def rotate(self, axis, angle, origin):
        """Rotate this geometry by a certain angle around an axis and origin.

        Args:
            axis: A ladybug_geometry Vector3D axis representing the axis of rotation.
            angle: An angle for rotation in degrees.
            origin: A ladybug_geometry Point3D for the origin around which the
                object will be rotated.
        """
        self._geometry = self.geometry.rotate(axis, math.radians(angle), origin)

    def rotate_xy(self, angle, origin):
        """Rotate this geometry counterclockwise in the world XY plane by an angle.

        Args:
            angle: An angle in degrees.
            origin: A ladybug_geometry Point3D for the origin around which the
                object will be rotated.
        """
        self._geometry = self.geometry.rotate_xy(math.radians(angle), origin)

    def reflect(self, plane):
        """Reflect this geometry across a plane.

        Args:
            plane: A ladybug_geometry Plane across which the object will
                be reflected.
        """
        self._geometry = self.geometry.reflect(plane.n, plane.o)

    def scale(self, factor, origin=None):
        """Scale this geometry by a factor from an origin point.

        Args:
            factor: A number representing how much the object should be scaled.
            origin: A ladybug_geometry Point representing the origin from which
                to scale. If None, it will be scaled from the World origin.
        """
        self._geometry = self.geometry.scale(factor, origin)

    def __repr__(self):
        return 'Ladybug Display 3D Base Object'


class _SingleColorBase3D(_DisplayBase3D):
    """A base class for ladybug-display geometry objects with a single color.

    Args:
        geometry: A ladybug-geometry object.
        color: A ladybug Color object. If None, a default black color will be
            used. (Default: None).

    Properties:
        * geometry
        * color
        * user_data
    """
    __slots__ = ('_color',)

    def __init__(self, geometry, color=None):
        """Initialize base with shade object."""
        _DisplayBase3D.__init__(self, geometry)
        self.color = color

    @property
    def color(self):
        """Get or set a color for this object."""
        return self._color

    @color.setter
    def color(self, value):
        if value is None:
            value = Color(0, 0, 0)
        else:
            assert isinstance(value, Color), 'Expected Color for ladybug_display ' \
                'object color. Got {}.'.format(type(value))
        self._color = value


class _SingleColorModeBase3D(_SingleColorBase3D):
    """A base class for ladybug-display geometry objects with a display mode.

    Args:
        geometry: A ladybug-geometry object.
        color: A ladybug Color object. If None, a default black color will be
            used. (Default: None).
        display_mode: Text to indicate the display mode (surface, wireframe, etc.).
            Choose from the following. (Default: Surface).

    Properties:
        * geometry
        * color
        * display_mode
        * user_data
    """
    __slots__ = ('_display_mode',)

    def __init__(self, geometry, color=None, display_mode='Surface'):
        """Initialize object."""
        _SingleColorBase3D.__init__(self, geometry, color)
        self.display_mode = display_mode

    @property
    def display_mode(self):
        """Get or set text to indicate the display mode."""
        return self._display_mode

    @display_mode.setter
    def display_mode(self, value):
        clean_input = value.lower()
        for key in DISPLAY_MODES:
            if key.lower() == clean_input:
                value = key
                break
        else:
            raise ValueError(
                'display_mode {} is not recognized.\nChoose from the '
                'following:\n{}'.format(value, DISPLAY_MODES))
        self._display_mode = value


class _LineCurveBase3D(_SingleColorBase3D):
    """A base class for all line-like 3D geometry objects.

    Args:
        geometry: A ladybug-geometry object.
        color: A ladybug Color object. If None, a default black color will be
            used. (Default: None).
        line_width: Number for line width in pixels (for the screen) or  millimeters
            (in print). This can also be the Default object to indicate that the
            default settings of the interface should be used.
        line_type: Get or set text to indicate the type of line to display.
            Choose from the following. (Default: "Continuous")

            * Continuous
            * Dashed
            * Dotted
            * DashDot
            * Center
            * Border
            * Hidden

    Properties:
        * geometry
        * color
        * line_width
        * line_type
        * user_data
    """
    __slots__ = ('_line_width', '_line_type')

    def __init__(self, geometry, color=None, line_width=default, line_type='Continuous'):
        """Initialize base with shade object."""
        _SingleColorBase3D.__init__(self, geometry, color)
        self.line_width = line_width
        self.line_type = line_type

    @property
    def line_width(self):
        """Get or set a color for this object."""
        return self._line_width

    @line_width.setter
    def line_width(self, value):
        if value == default:
            self._line_width = default
        else:
            self._line_width = float_positive(value, 'line width')

    @property
    def line_type(self):
        """Get or set text to indicate the type of line to display."""
        return self._line_type

    @line_type.setter
    def line_type(self, value):
        clean_input = value.lower()
        for key in LINE_TYPES:
            if key.lower() == clean_input:
                value = key
                break
        else:
            raise ValueError(
                'line_type {} is not recognized.\nChoose from the '
                'following:\n{}'.format(value, LINE_TYPES))
        self._line_type = value



# coding=utf-8
from ladybug_geometry.geometry2d.pointvector import Point2D, Vector2D
from ladybug_geometry.geometry2d.ray import Ray2D
from ladybug.color import Color
from ladybug_display.geometry2d.ray import DisplayRay2D


def test_display_ray2d_init():
    """Test the initialization of Ray2D objects and basic properties."""
    grey = Color(100, 100, 100)
    pt = Point2D(2, 0)
    vec = Vector2D(0, 2)
    ray = DisplayRay2D(Ray2D(pt, vec), grey)
    str(ray)  # test the string representation of the ray

    assert ray.color == grey
    assert ray.p == Point2D(2, 0)
    assert ray.v == Vector2D(0, 2)

    ray.reverse()
    assert ray.p == Point2D(2, 0)
    assert ray.v == Vector2D(0, -2)


def test_ray3_to_from_dict():
    """Test the to/from dict of Ray2D objects."""
    grey = Color(100, 100, 100)
    pt = Point2D(2, 0)
    vec = Vector2D(0, 2)
    ray = DisplayRay2D(Ray2D(pt, vec), grey)
    ray_dict = ray.to_dict()
    new_ray = DisplayRay2D.from_dict(ray_dict)
    assert isinstance(new_ray, DisplayRay2D)
    assert new_ray.to_dict() == ray_dict



"""Collection of methods for type input checking."""
import re
import math

try:
    INFPOS = math.inf
    INFNEG = -1 * math.inf
except AttributeError:
    # python 2
    INFPOS = float('inf')
    INFNEG = float('-inf')


def valid_string(value, input_name=''):
    """Check that a string is valid as an identifier."""
    try:
        illegal_match = re.search(r'[^.A-Za-z0-9_-]', value)
    except TypeError:
        raise TypeError('Input {} must be a text string. Got {}: {}.'.format(
            input_name, type(value), value))
    assert illegal_match is None, 'Illegal character "{}" found in {}'.format(
        illegal_match.group(0), input_name)
    assert len(value) > 0, 'Input {} "{}" contains no characters.'.format(
        input_name, value)
    assert len(value) <= 100, 'Input {} "{}" must be less than 100 characters.'.format(
        input_name, value)
    return value


def _number_check(value, input_name):
    """Check if value is a number."""
    try:
        number = float(value)
    except (ValueError, TypeError):
        raise TypeError('Input {} must be a number. Got {}: {}.'.format(
            input_name, type(value), value))
    return number


def float_in_range(value, mi=INFNEG, ma=INFPOS, input_name=''):
    """Check a float value to be between minimum and maximum."""
    number = _number_check(value, input_name)
    assert mi <= number <= ma, 'Input number {} must be between {} and {}. ' \
        'Got {}'.format(input_name, mi, ma, value)
    return number


def float_in_range_excl(value, mi=INFNEG, ma=INFPOS, input_name=''):
    """Check a float value to be greater than minimum and less than maximum."""
    number = _number_check(value, input_name)
    assert mi < number < ma, 'Input number {} must be greater than {} ' \
        'and less than {}. Got {}'.format(input_name, mi, ma, value)
    return number


def float_in_range_excl_incl(value, mi=INFNEG, ma=INFPOS, input_name=''):
    """Check a float value to be greater than minimum and less than/equal to maximum."""
    number = _number_check(value, input_name)
    assert mi < number <= ma, 'Input number {} must be greater than {} and less than ' \
        'or equal to {}. Got {}'.format(input_name, mi, ma, value)
    return number


def float_in_range_incl_excl(value, mi=INFNEG, ma=INFPOS, input_name=''):
    """Check a float value to be greater than/equal to minimum and less than maximum."""
    number = _number_check(value, input_name)
    assert mi <= number < ma, 'Input number {} must be greater than or equal to {} ' \
        'and less than {}. Got {}'.format(input_name, mi, ma, value)
    return number


def int_in_range(value, mi=INFNEG, ma=INFPOS, input_name=''):
    """Check an integer value to be between minimum and maximum."""
    try:
        number = int(value)
    except ValueError:
        # try to convert to float and then digit if possible
        try:
            number = int(float(value))
        except (ValueError, TypeError):
            raise TypeError('Input {} must be an integer. Got {}: {}.'.format(
                input_name, type(value), value))
    except (ValueError, TypeError):
        raise TypeError('Input {} must be an integer. Got {}: {}.'.format(
            input_name, type(value), value))
    assert mi <= number <= ma, 'Input integer {} must be between {} and {}. ' \
        'Got {}.'.format(input_name, mi, ma, value)
    return number


def float_positive(value, input_name=''):
    """Check a float value to be positive."""
    return float_in_range_excl(value, 0, INFPOS, input_name)


def int_positive(value, input_name=''):
    """Check if an integer value is positive."""
    return int_in_range(value, 0, INFPOS, input_name)


def tuple_with_length(value, length=3, item_type=float, input_name=''):
    """Try to create a tuple with a certain value."""
    try:
        value = tuple(item_type(v) for v in value)
    except (ValueError, TypeError):
        raise TypeError('Input {} must be a {}.'.format(
            input_name, item_type))
    assert len(value) == length, 'Input {} length must be {} not {}'.format(
        input_name, length, len(value))
    return value


def list_with_length(value, length=3, item_type=float, input_name=''):
    """Try to create a list with a certain value."""
    try:
        value = [item_type(v) for v in value]
    except (ValueError, TypeError):
        raise TypeError('Input {} must be a {}.'.format(
            input_name, item_type))
    assert len(value) == length, 'Input {} length must be {} not {}'.format(
        input_name, length, len(value))
    return value



# coding: utf-8
"""Base class for all geometry objects."""
LINE_TYPES = ('Continuous', 'Dashed', 'Dotted', 'DashDot', 'Center', 'Border', 'Hidden')
DISPLAY_MODES = ('Surface', 'SurfaceWithEdges', 'Wireframe', 'Points')


class _DisplayBase(object):
    """A base class for all ladybug-display geometry objects.

    Args:
        geometry: A ladybug-geometry object.

    Properties:
        * geometry
        * user_data
    """
    __slots__ = ('_geometry', '_user_data')

    def __init__(self, geometry):
        """Initialize base object."""
        self._geometry = geometry
        self._user_data = None

    @property
    def geometry(self):
        """Get a ladybug_geometry object."""
        return self._geometry

    @property
    def user_data(self):
        """Get or set an optional dictionary for additional meta data for this object.

        This will be None until it has been set. All keys and values of this
        dictionary should be of a standard Python type to ensure correct
        serialization of the object to/from JSON (eg. str, float, int, list, dict)
        """
        return self._user_data

    @user_data.setter
    def user_data(self, value):
        if value is not None:
            assert isinstance(value, dict), 'Expected dictionary for honeybee ' \
                'object user_data. Got {}.'.format(type(value))
        self._user_data = value

    def duplicate(self):
        """Get a copy of this object."""
        return self.__copy__()

    def __copy__(self):
        new_obj = self.__class__(self.geometry)
        new_obj._user_data = None if self.user_data is None else self.user_data.copy()
        return new_obj

    def ToString(self):
        return self.__repr__()

    def __repr__(self):
        return 'Ladybug Display Base Object'



"""A plane that can be displayed in 3D space."""
from ladybug_geometry.geometry3d.plane import Plane
from ladybug.color import Color

from ._base import _SingleColorBase3D


class DisplayPlane(_SingleColorBase3D):
    """A plane in 3D space with display properties.

    Args:
        geometry: A ladybug-geometry Plane object.
        color: A ladybug Color object. If None, a default black color will be
            used. (Default: None).
        show_axes: A boolean to note whether the plane should be displayed with
            XY axes instead of just an origin point and a normal vector.
        show_grid: A boolean to note whether the plane should be displayed
            with a grid.

    Properties:
        * geometry
        * color
        * show_axes
        * show_grid
        * o
        * n
        * x
        * y
        * k
        * user_data
    """
    __slots__ = ('_show_axes', '_show_grid')

    def __init__(self, geometry, color=None, show_axes=False, show_grid=False):
        """Initialize base with shade object."""
        assert isinstance(geometry, Plane), '\
            Expected ladybug_geometry Plane. Got {}'.format(type(geometry))
        _SingleColorBase3D.__init__(self, geometry, color)
        self.show_axes = show_axes
        self.show_grid = show_grid

    @classmethod
    def from_dict(cls, data):
        """Initialize a DisplayPlane from a dictionary.

        Args:
            data: A dictionary representation of an DisplayPlane object.
        """
        assert data['type'] == 'DisplayPlane', \
            'Expected DisplayPlane dictionary. Got {}.'.format(data['type'])
        color = Color.from_dict(data['color']) if 'color' in data and data['color'] \
            is not None else None
        ax = data['show_axes'] if 'show_axes' in data else False
        gd = data['show_grid'] if 'show_grid' in data else False
        geo = cls(Plane.from_dict(data['geometry']), color, ax, gd)
        if 'user_data' in data and data['user_data'] is not None:
            geo.user_data = data['user_data']
        return geo

    @property
    def show_axes(self):
        """Get or set a boolean for whether to display the XY axes of the plane."""
        return self._show_axes

    @show_axes.setter
    def show_axes(self, value):
        try:
            self._show_axes = bool(value)
        except TypeError:
            raise TypeError(
                'Expected boolean for DisplayPlane.show_axes. Got {}.'.format(value))

    @property
    def show_grid(self):
        """Get or set a boolean for whether to display the grid of the plane."""
        return self._show_grid

    @show_grid.setter
    def show_grid(self, value):
        try:
            self._show_grid = bool(value)
        except TypeError:
            raise TypeError(
                'Expected boolean for DisplayPlane.show_grid. Got {}.'.format(value))

    @property
    def o(self):
        """Get a Point3D representing the origin of the plane."""
        return self._geometry.o

    @property
    def n(self):
        """Get a Vector3D representing the normal of the plane."""
        return self._geometry.n

    @property
    def x(self):
        """Get  aVector3D representing the x axis of the plane."""
        return self._geometry.x

    @property
    def y(self):
        """Get a Vector3D representing the y axis of the plane."""
        return self._geometry.y

    @property
    def k(self):
        """Get a number for the constant of the plane."""
        return self._geometry.k

    def to_dict(self):
        """Return DisplayPlane as a dictionary."""
        base = {'type': 'DisplayPlane'}
        base['geometry'] = self._geometry.to_dict()
        base['color'] = self.color.to_dict()
        base['show_axes'] = self.show_axes
        base['show_grid'] = self.show_grid
        if self.user_data is not None:
            base['user_data'] = self.user_data
        return base

    def __copy__(self):
        new_g = DisplayPlane(self.geometry, self.color, self.show_axes, self.show_grid)
        new_g._user_data = None if self.user_data is None else self.user_data.copy()
        return new_g

    def __repr__(self):
        return 'DisplayPlane: {}'.format(self.geometry)



"""Geometry objects for display in 3D space."""

from .vector import DisplayVector3D
from .point import DisplayPoint3D
from .ray import DisplayRay3D
from .line import DisplayLineSegment3D
from .arc import DisplayArc3D
from .polyline import DisplayPolyline3D
from .plane import DisplayPlane
from .face import DisplayFace3D
from .mesh import DisplayMesh3D
from .polyface import DisplayPolyface3D
from .sphere import DisplaySphere
from .cone import DisplayCone
from .cylinder import DisplayCylinder



# coding=utf-8
from ladybug_geometry.geometry3d.pointvector import Point3D, Vector3D
from ladybug_geometry.geometry3d.cylinder import Cylinder
from ladybug.color import Color
from ladybug_display.geometry3d.cylinder import DisplayCylinder


def test_cylinder_init():
    """Test the initialization of Cylinder objects and basic properties."""
    grey = Color(100, 100, 100)
    center = Point3D(2, 0, 2)
    axis = Vector3D(0, 2, 2)
    radius = 0.7
    c = DisplayCylinder(Cylinder(center, axis, radius), grey)
    str(c)  # test the string representation of the cylinder

    assert c.color == grey
    assert c.display_mode == 'Surface'
    assert c.center == Point3D(2, 0, 2)
    assert c.axis == Vector3D(0, 2, 2)
    assert c.radius == 0.7
    assert c.height == c.axis.magnitude
    assert isinstance(c.area, float)
    assert isinstance(c.volume, float)

    blue = Color(0, 0, 100)
    c.color = blue
    c.display_mode = 'Wireframe'
    assert c.color == blue
    assert c.display_mode == 'Wireframe'


def test_cylinder_to_from_dict():
    """Test the Cone to_dict and from_dict methods."""
    grey = Color(100, 100, 100)
    center = Point3D(2, 0, 2)
    axis = Vector3D(0, 2, 2)
    radius = 0.7
    c = DisplayCylinder(Cylinder(center, axis, radius), grey)
    con_d = c.to_dict()
    new_c = DisplayCylinder.from_dict(con_d)
    assert isinstance(new_c, DisplayCylinder)
    assert new_c.to_dict() == con_d
